# Copyright 2025 Samarth Mistry
# This file is part of the `opti-extensions` package, which is released under
# the Apache Licence, Version 2.0 (http://www.apache.org/licenses/LICENSE-2.0).

"""Functions to add variable(s) to a DOcplex model."""

from __future__ import annotations

from collections.abc import Callable, Sequence
from typing import Literal, overload

from docplex.mp.dvar import Var
from docplex.mp.model import Model
from docplex.mp.vartype import (
    BinaryVarType,
    ContinuousVarType,
    IntegerVarType,
    SemiContinuousVarType,
    SemiIntegerVarType,
)

from .._index_sets import Elem1DT, ElemNDT, IndexSet1D, IndexSetND
from .._param_dicts import ParamDict1D, ParamDictND, ParamT
from ._var_dicts import VarDict1D, VarDictND


def _get_docpx_vartype(
    vartype: str, model: Model
) -> (
    ContinuousVarType | BinaryVarType | IntegerVarType | SemiContinuousVarType | SemiIntegerVarType
):
    """Get the DOcplex variable type from the corresponding string name.

    Parameters
    ----------
    vartype : str
        Variable type.
    model : docplex.mp.model.Model
        DOcplex model.

    Returns
    -------
    DOcplex variable type (subclass of docplex.mp.vartype.VarType)

    Raises
    ------
    ValueError
        If the variable type name is invalid.
    """
    if not isinstance(vartype, str):
        raise TypeError('`vartype` should be a string')

    match vartype.lower():
        case 'continuous' | 'c':
            return model.continuous_vartype
        case 'binary' | 'b':
            return model.binary_vartype
        case 'integer' | 'i':
            return model.integer_vartype
        case 'semicontinuous' | 'sc':
            return model.semicontinuous_vartype
        case 'semiinteger' | 'si':
            return model.semiinteger_vartype
        case _:
            raise ValueError('`vartype` is invalid')


def _paramdictNd_as_bound(
    indexset: IndexSetND[ElemNDT],
    bound: ParamDictND[ElemNDT, ParamT],
    bound_type: Literal['lb', 'ub'],
) -> Callable[..., int | float | None]:
    """Use ParamDictND as DOcplex variable bound.

    Parameters
    ----------
    indexset : IndexSetND
        Index-set for defining the variables.
    bound : ParamDictND
        ParamDict to be used as variable bound.
    bound_type : str
        Bound type, either ``'lb'`` or ``'ub'``.

    Returns
    -------
    function
        ParamDictND.get method that can be used as DOcplex variable bound for the
        `Model.xyz_var_dict` method.

    Raises
    ------
    ValueError
        ParamDict keys and indexset have tuple elements of different lengths.
    """
    if (not bound) or (bound and bound._indexset._tuplelen == indexset._tuplelen):
        return bound.get
    else:
        raise ValueError(f'{bound_type} keys and indexset have tuple elements of different lengths')


def _preprocess_bound(
    indexset: IndexSet1D[Elem1DT] | IndexSetND[ElemNDT],
    bound: int
    | float
    | Sequence[int | float | None]
    | Callable[..., int | float | None]
    | ParamDict1D[Elem1DT, ParamT]
    | ParamDictND[ElemNDT, ParamT]
    | None,
    bound_type: Literal['lb', 'ub'],
) -> int | float | Sequence[int | float | None] | Callable[..., int | float | None] | None:
    """Preprocess DOcplex variable bound.

    Parameters
    ----------
    indexset : IndexSet1D or IndexSetND
        Index-set for defining the variables.
    bound : int or float or Sequence or function or ParamDict or None
        Variable bound.
    bound_type : str
        Bound type, either ``'lb'`` or ``'ub'``.

    Returns
    -------
    int or float or Sequence or function or None
        DOcplex variable bound for the `Model.xyz_var_dict` method.

    Raises
    ------
    ValueError
        ParamDict keys and indexset have tuple elements of different lengths.
    TypeError
        If paramdict is ParamDictND and indexset is IndexSet1D, or vice versa.
    """
    # Validate if ParamDict is used as variable bound
    if isinstance(bound, ParamDict1D):
        if isinstance(indexset, IndexSet1D):
            return bound.get
        else:
            raise TypeError(f'`{bound_type}` should be ParamDictND when indexset is IndexSetND')
    elif isinstance(bound, ParamDictND):
        if isinstance(indexset, IndexSetND):
            return _paramdictNd_as_bound(indexset, bound, bound_type)
        else:
            raise TypeError(f'`{bound_type}` should be ParamDict1D when indexset is IndexSet1D')
    # Let DOcplex handle everything else, so return as is
    else:
        return bound


@overload
def add_variable(  # numpydoc ignore=GL08
    model: Model,
    vartype: Literal[
        'continuous',
        'integer',
        'binary',
        'semicontinuous',
        'semiinteger',
        'C',
        'I',
        'B',
        'SC',
        'SI',
    ],
    lb: int | float,
    ub: int | float | None = ...,
    name: str | None = ...,
) -> Var: ...


@overload
def add_variable(  # numpydoc ignore=GL08
    model: Model,
    vartype: Literal['continuous', 'integer', 'binary', 'C', 'I', 'B'],
    lb: int | float | None = ...,
    ub: int | float | None = ...,
    name: str | None = ...,
) -> Var: ...


def add_variable(
    model: Model,
    vartype: Literal[
        'continuous',
        'integer',
        'binary',
        'semicontinuous',
        'semiinteger',
        'C',
        'I',
        'B',
        'SC',
        'SI',
    ],
    lb: int | float | None = None,
    ub: int | float | None = None,
    name: str | None = None,
) -> Var:
    """Create and add a single variable to the DOcplex model.

    This function is a simple wrapper around DOcplex's `xyz_var` methods and provided for a
    consistent API experience.

    Parameters
    ----------
    model : docplex.mp.model.Model
        DOcplex model.
    vartype : str
        Variable type, in one of the following forms:

        * ``'continuous'`` or ``'C'``
        * ``'binary'`` or ``'B'``
        * ``'integer'`` or ``'I'``
        * ``'semicontinuous'`` or ``'SC'``
        * ``'semiinteger'`` or ``'SI'``

    lb : int or float
        Lower bound. Default ``None`` corresponds to:

        * Continuous: 0
        * Binary: 0
        * Integer: 0
        * Semicontinuous: not applicable... will raise ValueError
        * Semiinteger: not applicable... will raise ValueError

    ub : int or float, optional
        Upper bound. Default ``None`` corresponds to:

        * Continuous: infinity (1e+20)
        * Binary: 1
        * Integer: infinity (1e+20)
        * Semicontinuous: infinity (1e+20)
        * Semiinteger: infinity (1e+20)

    name : str, optional
        Variable name, by default ``None``.

    Returns
    -------
    docplex.mp.dvar.Var

    Raises
    ------
    ValueError
        If the variable type is invalid.
    ValueError
        If no lower bound is specified for Semicontinuous or Semiinteger variable type.

    See Also
    --------
    add_variables : For multiple variables (corresponding to an index-set).

    Examples
    --------
    Create DOcplex model:

    >>> from docplex.mp.model import Model
    >>> mdl = Model()

    Add variables:

    >>> x = add_variable(mdl, 'C', name='x')
    >>> x
    docplex.mp.Var(type=C,name='x')

    >>> y = add_variable(mdl, 'semicontinuous', lb=2, ub=5, name='y')
    >>> y
    docplex.mp.Var(type=S,name='y',lb=2,ub=5)
    """
    if not isinstance(model, Model):
        raise TypeError('`model` should be docplex.mp.model.Model')

    vt = _get_docpx_vartype(vartype, model)
    if vartype.lower() in ('semicontinuous', 'sc') and lb is None:
        raise ValueError('Need to set a lower bound for Semicontinuous variable type')
    if vartype.lower() in ('semiinteger', 'si') and lb is None:
        raise ValueError('Need to set a lower bound for Semiinteger variable type')

    return model.var(vt, lb, ub, name)


@overload
def add_variables(  # numpydoc ignore=GL08
    model: Model,
    indexset: IndexSet1D[Elem1DT],
    vartype: Literal[
        'continuous',
        'integer',
        'binary',
        'semicontinuous',
        'semiinteger',
        'C',
        'I',
        'B',
        'SC',
        'SI',
    ],
    lb: int
    | float
    | Sequence[int | float]
    | Callable[..., int | float]
    | ParamDict1D[Elem1DT, ParamT],
    ub: int
    | float
    | Sequence[int | float | None]
    | Callable[..., int | float | None]
    | ParamDict1D[Elem1DT, ParamT]
    | None = ...,
    name: str | Callable[..., str] | None = ...,
    key_format: str | None = ...,
) -> VarDict1D[Elem1DT, Var]: ...


@overload
def add_variables(  # numpydoc ignore=GL08
    model: Model,
    indexset: IndexSet1D[Elem1DT],
    vartype: Literal['continuous', 'integer', 'binary', 'C', 'I', 'B'],
    lb: int
    | float
    | Sequence[int | float | None]
    | Callable[..., int | float | None]
    | ParamDict1D[Elem1DT, ParamT]
    | None = ...,
    ub: int
    | float
    | Sequence[int | float | None]
    | Callable[..., int | float | None]
    | ParamDict1D[Elem1DT, ParamT]
    | None = ...,
    name: str | Callable[..., str] | None = ...,
    key_format: str | None = ...,
) -> VarDict1D[Elem1DT, Var]: ...


@overload
def add_variables(  # numpydoc ignore=GL08
    model: Model,
    indexset: IndexSetND[ElemNDT],
    vartype: Literal[
        'continuous',
        'integer',
        'binary',
        'semicontinuous',
        'semiinteger',
        'C',
        'I',
        'B',
        'SC',
        'SI',
    ],
    lb: int
    | float
    | Sequence[int | float]
    | Callable[..., int | float]
    | ParamDictND[ElemNDT, ParamT],
    ub: int
    | float
    | Sequence[int | float | None]
    | Callable[..., int | float | None]
    | ParamDictND[ElemNDT, ParamT]
    | None = ...,
    name: str | Callable[..., str] | None = ...,
    key_format: str | None = ...,
) -> VarDictND[ElemNDT, Var]: ...


@overload
def add_variables(  # numpydoc ignore=GL08
    model: Model,
    indexset: IndexSetND[ElemNDT],
    vartype: Literal['continuous', 'integer', 'binary', 'C', 'I', 'B'],
    lb: int
    | float
    | Sequence[int | float | None]
    | Callable[..., int | float | None]
    | ParamDictND[ElemNDT, ParamT]
    | None = ...,
    ub: int
    | float
    | Sequence[int | float | None]
    | Callable[..., int | float | None]
    | ParamDictND[ElemNDT, ParamT]
    | None = ...,
    name: str | Callable[..., str] | None = ...,
    key_format: str | None = ...,
) -> VarDictND[ElemNDT, Var]: ...


def add_variables(
    model: Model,
    indexset: IndexSet1D[Elem1DT] | IndexSetND[ElemNDT],
    vartype: Literal[
        'continuous',
        'integer',
        'binary',
        'semicontinuous',
        'semiinteger',
        'C',
        'I',
        'B',
        'SC',
        'SI',
    ],
    lb: int
    | float
    | Sequence[int | float | None]
    | Callable[..., int | float | None]
    | ParamDict1D[Elem1DT, ParamT]
    | ParamDictND[ElemNDT, ParamT]
    | None = None,
    ub: int
    | float
    | Sequence[int | float | None]
    | Callable[..., int | float | None]
    | ParamDict1D[Elem1DT, ParamT]
    | ParamDictND[ElemNDT, ParamT]
    | None = None,
    name: str | Callable[..., str] | None = None,
    key_format: str | None = None,
) -> VarDict1D[Elem1DT, Var] | VarDictND[ElemNDT, Var]:
    """Create and add multiple variables (corresponding to an index-set) to the DOcplex model.

    This function extends DOcplex's `Model.xyz_var_dict` methods by wrapping them and returning
    VarDict1D/VarDictND data structures - that can efficiently sum subsets of variables with
    a user-friendly syntax based on wildcard patterns.

    Parameters
    ----------
    model : docplex.mp.model.Model
        DOcplex model.
    indexset : IndexSet1D or IndexSetND
        Index-set for defining the variables. The ``name``/``names`` attribute of
        IndexSet1D/IndexSetND will be set as the ``key_name``/``key_names`` attribute of
        VarDict1D/VarDictND.
    vartype : str
        Variable type, in one of the following forms:

        * ``'continuous'`` or ``'C'``
        * ``'binary'`` or ``'B'``
        * ``'integer'`` or ``'I'``
        * ``'semicontinuous'`` or ``'SC'``
        * ``'semiinteger'`` or ``'SI'``

    lb : int or float or sequence or function or ParamDict, optional
        Lower bound, in one of the following forms:

        * A number - if all variables share the same lower bound.
        * A sequence of numbers - one for each variable.
        * A function - that returns a number when called on each element of index-set.
        * A ParamDict - with keys following the same structute as the index-set elements and values
          representing the lower bound; will fallback to ``None`` for index-set elements not found
          in ParamDict keys.

        Default ``None`` corresponds to:

        * Continuous: 0
        * Binary: 0
        * Integer: 0
        * Semicontinuous: not applicable... will raise ValueError
        * Semiinteger: not applicable... will raise ValueError

    ub : int or float or sequence or function or ParamDict, optional
        Upper bound, in one of the following forms:

        * A number - if all variables share the same lower bound.
        * A sequence of numbers - one for each variable.
        * A function - that returns a number when called on each element of the index-set.
        * A ParamDict - with keys following the same structute as the index-set elements and values
          representing the upper bound; will fallback to ``None`` for index-set elements not found
          in ParamDict keys.

        Default ``None`` corresponds to ``1`` for binary variable type and ``infinity`` for other
        types.
    name : str or function, optional
        For naming variables, in one of the following forms:

        * A string - applied as a prefix to the string representation of each element of the
          index-set.
        * A function - that generates a name when called on each element of the index-set.
        * ``None``.

        Default is None. It will also be used as the ``value_name`` attribute of
        VarDict1D/VarDictND.
    key_format : format str (should include '%s'), optional
        Defines how index-set elements are incorporated into variable names, by default ``'_%s'``.

    Returns
    -------
    VarDict1D or VarDictND

    Raises
    ------
    ValueError
        If the index-set is empty.
    ValueError
        If the variable type is invalid.
    ValueError
        If no lower bound is specified for Semicontinuous or Semiinteger variable type.

    See Also
    --------
    add_variable : For a single variable.

    Examples
    --------
    Create DOcplex model:

    >>> from docplex.mp.model import Model
    >>> mdl = Model()

    Create index-sets:

    >>> nodes = IndexSet1D(['A', 'B', 'C'], name='node')
    >>> arcs = IndexSetND([('A', 'B'), ('B', 'C'), ('C', 'B')], names=['ori', 'des'])

    Add variables:

    >>> node_select = add_variables(mdl, nodes, 'B', name='select')
    >>> node_select
    VarDict1D: node -> select
    {'A': docplex.mp.Var(type=B,name='select_A'),
     'B': docplex.mp.Var(type=B,name='select_B'),
     'C': docplex.mp.Var(type=B,name='select_C')}

    >>> arc_flow = add_variables(mdl, arcs, 'C', ub=10, name='flow')
    >>> arc_flow
    VarDictND: (ori, des) -> flow
    {('A', 'B'): docplex.mp.Var(type=C,name='flow_A_B',ub=10),
     ('B', 'C'): docplex.mp.Var(type=C,name='flow_B_C',ub=10),
     ('C', 'B'): docplex.mp.Var(type=C,name='flow_C_B',ub=10)}
    """
    if not isinstance(model, Model):
        raise TypeError('`model` should be docplex.mp.model.Model')
    if not isinstance(indexset, IndexSet1D | IndexSetND):
        raise TypeError('`indexset` should be either IndexSet1D or IndexSetND')
    if not indexset:
        raise ValueError(f'{indexset.__class__.__name__} is empty')

    lb = _preprocess_bound(indexset, lb, 'lb')
    ub = _preprocess_bound(indexset, ub, 'ub')

    vt = _get_docpx_vartype(vartype, model)
    if vt.is_semi_type():
        # Provide informative error messages since those from DOcplex are not easy to understand
        if lb is None:
            raise ValueError(f'Need to set a lower bound for {vt.short_name} variable type')
        if (isinstance(lb, Sequence) and None in lb) or (
            callable(lb) and any(lb(elem) is None for elem in indexset)
        ):
            raise ValueError(
                f'Did not get a lower bound for all variables; not allowed for {vt.short_name}'
                'variable type'
            )

    docpx_var_dict = model.var_dict(indexset, vt, lb, ub, name, key_format)
    value_name = name if isinstance(name, str) else None

    if isinstance(indexset, IndexSet1D):
        return VarDict1D._create(
            docpx_var_dict, indexset, model=model, vartype=vt, value_name=value_name
        )
    return VarDictND._create(
        docpx_var_dict, indexset, model=model, vartype=vt, value_name=value_name
    )



# Copyright 2025 Samarth Mistry
# This file is part of the `opti-extensions` package, which is released under
# the Apache Licence, Version 2.0 (http://www.apache.org/licenses/LICENSE-2.0).

"""DOcplex model-specific functions."""

from io import TextIOBase
from typing import Any, Literal, TextIO

from docplex.mp.model import Model
from docplex.mp.solution import SolveSolution


def print_problem_stats(model: Model) -> None:
    """Print problem statistics of a DOcplex model.

    Parameters
    ----------
    model : docplex.mp.model.Model
        DOcplex model.

    See Also
    --------
    print_solution_quality_stats : For printing quality statistics of the incumbent solution.
    """
    cplex = model.get_cplex()
    if cplex.get_problem_name() != model.name:
        cplex.set_problem_name(model.name)
    print(cplex.get_stats())


def print_solution_quality_stats(model: Model) -> None:
    """Print quality statistics of the incumbent solution of a DOcplex model, if available.

    Parameters
    ----------
    model : docplex.mp.model.Model
        DOcplex model.

    See Also
    --------
    print_problem_stats : For printing problem statistics.
    """
    if model.solution is None:
        print(f'Model `{model.name}` has no incumbent solution.')
    else:
        print(model.get_cplex().solution.get_quality_metrics())


def solve(model: Model, **kwargs: Any) -> SolveSolution | None:
    """Solve a DOcplex model with additional logging output for problem and solution statistics.

    This function is a simple wrapper around `docplex.mp.model.Model`'s `solve` method and takes the
    same keyword arguments.

    Parameters
    ----------
    model : docplex.mp.model.Model
        DOcplex model.
    log_output : bool or str or stream object, optional
        Log output switch, in one of the following forms:

        * ``True`` or ``'1'`` or ``'stdout'`` or ``'sys.stdout'``: Log is output to stdout.
        * ``'stderr'`` or ``'sys.stderr'``: Log is output to stderr.
        * ``False`` or ``'0'`` or ``None``: No log output.
        * File path (in form of str): Log is output to the file.
        * Stream object (a file-like object with a write method and a flush method): Log is output
          to the stream object.

        Default is the value of `context.solver.log_output` setting. Specifying this argument
        overwrites that setting.
    clean_before_solve : bool, optional
        By default, `solve` typically resumes from the point where the previous optimizer solve
        ended, if any. Setting this flag to ``True`` forces a clean restart.
    context : docplex.mp.context.Context, optional
        Context instance to be used in instead of the context the model was built with.
    cplex_parameters : docplex.mp.params.parameters.RootParameterGroup or dict, optional:
        A set of CPLEX parameters to use instead of the parameters defined as
        `context.cplex_parameters`, in one of the following forms:

        * `RootParameterGroup`: This can obtained by cloning the model's parameters
        * `dict` of parameter names and values.

    checker : str, optional
        Type of checks performed, in one of the following forms:

        * ``'std'``: Perform type checks on arguments to methods, check whether numerical arguments
          are valid numbers (but no check for NaN and infinity).
        * ``'numeric'``: Check whether numerical arguments are valid numbers, and not NaN or
          infinity.
        * ``'full'``: Perform all possible checks (combination of ``'std'`` and ``'numeric'``).
        * ``'off'``: Perform no checks at all. Disabling all checks might speed up model build time,
          but recommended only when the model has been thoroughly tested.

        Default is ``'std'``.
    parameter_sets : Iterable[ParameterSet], optional:
        An iterable of `ParameterSet` to be used with multi-objective optimization only. See the
        `create_parameter_sets` method of `docplex.mp.model.Model`.

    Returns
    -------
    docplex.mp.solution.SolveSolution or None
        DOcplex model solution if the optimizer managed to find a feasible solution, else None.

    See Also
    --------
    print_problem_stats : For printing problem statistics separately.
    print_solution_quality_stats : For printing quality statistics of the incumbent solution
        separately.
    """
    if not isinstance(model, Model):
        raise TypeError('`model` should be docplex.mp.model.Model')

    if 'log_output' in kwargs:  # kwargs take precedence
        out = kwargs['log_output']
    elif model.context.solver.log_output:  # then current context
        out = model.context.solver.log_output_as_stream
    else:  # nothing otherwise
        out = False
    to_log = out not in (False, '0', None)

    if to_log:
        div_len = 85
        cplex = model.get_cplex()

        # Get the output stream that will be used by DOcplex
        model.context.update_key_value('log_output', out)
        stream = model.context.solver.log_output_as_stream
        # Flag if the stream will be auto-closed by `model.solve`, and hence need to be reopened
        to_reopen = True if hasattr(stream, 'custom_close') else False

        # Write problem statistics
        if cplex.get_problem_name() != model.name:
            cplex.set_problem_name(model.name)
        log_header = f'  {model.problem_type} problem statistics  '.center(div_len, '-') + '\n\n'
        log_header += str(cplex.get_stats())
        log_header += '\n' + '  CPLEX optimizer log  '.center(div_len, '-') + '\n\n'
        stream.write(log_header)

    # Don't close the output stream; hand it over to `model.solve`
    solve_setting = stream if to_log else None
    # Remove if stream is in `kwargs` since we're handing it through an explicit argument
    kwargs.pop('log_output', None)
    # Invoke the actual method
    solution = model.solve(log_output=solve_setting, **kwargs)

    if to_log:
        # `model.solve` auto-closes stream objects, so reopen
        if to_reopen:
            stream = open(stream._target.name, 'a')

        # Write solution quality statistics if CPLEX finds a feasible solution
        if solution is None:
            stream.write('\n' + '-' * div_len + '\n')
        else:
            log_footer = '\n' + '  Solution quality statistics  '.center(div_len, '-') + '\n\n'
            log_footer += str(cplex.solution.get_quality_metrics())
            log_footer += '\n' + '-' * div_len + '\n'
            stream.write(log_footer)
        stream.flush()

        # When logging to stream objects, close them at the end
        if to_reopen:
            stream.close()

    return solution


def runseeds(
    model: Model,
    count: int = 30,
    log_output: Literal[True] | str | TextIO = True,
) -> None:
    """Evaluate variability of a mixed-integer DOcplex model by solving with different random seeds.

    This function is a wrapper around CPLEX's `runseeds` procedure that reports variability
    statistics in the output log. Refer to the CPLEX user manual for more details about this
    functionality.

    Parameters
    ----------
    model : docplex.mp.model.Model
        DOcplex model.
    count : int
        The number of times to solve the model, by default `30` (same as CPLEX).
    log_output : True or str or stream object, optional
        Log output switch, in one of the following forms:

        * ``True`` or ``'1'`` or ``'stdout'`` or ``'sys.stdout'``: Log is output to stdout.
        * ``'stderr'`` or ``'sys.stderr'``: Log is output to stderr.
        * File path (in form of str): Log is output to the file.
        * Stream object (a file-like object with a write method and a flush method): Log is output
          to the stream object.

        Default is True.

    Raises
    ------
    ValueError
        If the DOcplex model is not mixed-integer (MILP, MIQP, MIQCP).
    ValueError
        If count is not a positive number.
    """
    if not isinstance(model, Model):
        raise TypeError('`model` should be docplex.mp.model.Model')
    if model.problem_type not in ('MILP', 'MIQP', 'MIQCP'):
        raise ValueError('runseeds only supports mixed-integer DOcplex models (MILP, MIQP, MIQCP)')

    if not isinstance(count, int) or count < 1:
        raise ValueError('`count` should be a positive integer')

    if log_output in (False, '0', None):
        raise ValueError('`log_output` should be True or str or a stream object')
    elif log_output == '1':
        log_output = True
    if not (
        log_output is True
        or isinstance(log_output, str | TextIOBase)
        or (hasattr(log_output, 'write') and hasattr(log_output, 'flush'))
    ):
        raise TypeError('`log_output` should be True or str or a stream object')

    # Register prior state so it can be restored at the end
    prior_log_output = model.log_output

    # Set log_output stream
    model.log_output = log_output
    stream = model.context.solver.log_output_as_stream

    # Invoke the runseeds procedure
    model.apply_parameters()
    cplex = model.get_cplex()
    cplex.runseeds(cnt=count)

    # Close stream if it's a file created by DOcplex (when log_output='filepath')
    try:
        stream.custom_close()
    except AttributeError:
        pass

    # Restore prior state
    model.log_output = prior_log_output



# Copyright 2025 Samarth Mistry
# This file is part of the `opti-extensions` package, which is released under
# the Apache Licence, Version 2.0 (http://www.apache.org/licenses/LICENSE-2.0).

"""VarDict data structures."""

from __future__ import annotations

from collections.abc import Iterable
from typing import Any, Literal, TypeVar, cast

from docplex.mp.dvar import Var
from docplex.mp.linear import LinearExpr, ZeroExpr
from docplex.mp.model import Model
from docplex.mp.quad import QuadExpr
from docplex.mp.vartype import VarType

from .._dict_mixins import Dict1DMixin, DictNDMixin
from .._index_sets import Elem1DT, ElemNDT, ElemT, IndexSet1D, IndexSetBase, IndexSetND
from .._param_dicts import ParamDict1D, ParamDictND, ParamT
from .._var_dict_base import VarDictBase, _validate_for_dot_1d, _validate_for_dot_Nd

VarT = TypeVar('VarT', bound=Var)


class VarDictCore(VarDictBase[ElemT, VarT]):
    """Base class for custom subclasses of `dict` to define DOcplex model variables.

    Provides runtime type checks to ensure variables are `gurobipy.Var` objects.

    Parameters
    ----------
    docpx_var_dict : dict
        Dictionary of variable objects associated with the DOcplex model.
    indexset : IndexSetBase
        Keys of dictionary encapsulated in an IndexSet.
    model : docplex.mp.model.Model
        DOcplex model associated with the variable objects.
    vartype : docplex.mp.vartype.VarType
        Type corresponding to the variable objects.
    """

    __slots__ = ('_model', '_vartype')

    def __init__(
        self,
        docpx_var_dict: dict[ElemT, VarT],
        /,
        *,
        indexset: IndexSetBase[ElemT],
        model: Model,
        vartype: VarType,
    ) -> None:
        self._validate_docpx_var_dict(docpx_var_dict)

        self._model = model
        self._vartype = vartype

        super().__init__(docpx_var_dict, indexset)

    @property
    def model(self) -> Model:
        """The DOcplex model associated with the variables.

        Returns
        -------
        docplex.mp.model.Model
        """
        return self._model

    @property
    def vartype(self) -> VarType:
        """The DOcplex VarType corresponding to the variables.

        Returns
        -------
        DOcplex variable type (subclass of docplex.mp.vartype.VarType)
        """
        return self._vartype

    @staticmethod
    def _validate_docpx_var_dict(docpx_var_dict: dict[ElemT, VarT]) -> None:
        """Validate that the input is a populated dict of DOcplex variables.

        Parameters
        ----------
        docpx_var_dict : dict
            Input dictionary for VarDict.
        """
        if not isinstance(docpx_var_dict, dict):
            raise TypeError('input should be a dict')
        if not docpx_var_dict:
            raise ValueError('dict should be populated')
        if any(not isinstance(x, Var) for x in docpx_var_dict.values()):
            raise TypeError('dict values should be docplex.mp.dvar.Var objects')


class VarDict1D(VarDictCore[Elem1DT, VarT], Dict1DMixin[Elem1DT, VarT]):
    """Custom subclasses of `dict` to define DOcplex model variables with 1-dim scalar keys.

    Parameters
    ----------
    docpx_var_dict : dict
        Dictionary of variable objects from docplex.
    indexset : IndexSet1D
        Keys of dictionary encapsulated in IndexSet1D.
    model : docplex.mp.model.Model
        DOcplex model associated with the variable objects.
    vartype : docplex.mp.vartype.VarType
        DOcplex VarType corresponding to the variable objects.
    value_name : str, optional
        Name to refer to variables - not used internally, and solely for user reference.
    """

    # Private attributes
    # ------------------
    # _indexset : IndexSet1D
    #     Index-set of keys.

    __slots__ = ('_key_name', '_value_name')

    def __init__(
        self,
        docpx_var_dict: dict[Elem1DT, VarT],
        indexset: IndexSet1D[Elem1DT],
        /,
        *,
        model: Model,
        vartype: VarType,
        value_name: str | None = None,
    ) -> None:
        self.key_name = indexset.name
        self.value_name = value_name

        super().__init__(docpx_var_dict, indexset=indexset, model=model, vartype=vartype)

    def __new__(
        cls,
        docpx_var_dict: dict[Elem1DT, VarT],
        indexset: IndexSet1D[Elem1DT],
        /,
        *,
        model: Model,
        vartype: VarType,
        value_name: str | None = None,
    ) -> VarDict1D[Elem1DT, VarT]:
        raise TypeError(
            'This class is not meant to be instantiated; VarDict1D is built through the '
            'opti_extensions.docplex.add_variables function'
        )

    @classmethod
    def _create(
        cls,
        docpx_var_dict: dict[Elem1DT, VarT],
        indexset: IndexSet1D[Elem1DT],
        /,
        *,
        model: Model,
        vartype: VarType,
        value_name: str | None = None,
    ) -> VarDict1D[Elem1DT, VarT]:
        # Private method to construct VarDict1D
        instance = super().__new__(cls)
        cls.__init__(
            instance, docpx_var_dict, indexset, model=model, vartype=vartype, value_name=value_name
        )
        return instance

    def __repr__(self) -> str:
        # Printable string representation.
        return f'{self._get_repr_header()}\n{super().__repr__()}'

    def _repr_pretty_(self, p, cycle: bool) -> None:  # type: ignore[no-untyped-def]
        # Pretty repr for IPython.
        # https://ipython.readthedocs.io/en/stable/api/generated/IPython.lib.pretty.html#extending
        # Since IPython is not typed, we'll add a type ignore comment here
        # The annotation for arg `p` is `IPython.lib.pretty.RepresentationPrinter`
        p.text(f'{self._get_repr_header()}\n')
        p.pretty(dict(self))

    def lookup(self, key: Elem1DT) -> VarT | Literal[0]:
        """Get the variable for the specified key, or zero if it is not found.

        Parameters
        ----------
        key : key

        Returns
        -------
        docplex.mp.dvar.Var or ``0``

        Examples
        --------
        Create DOcplex model:

        >>> from docplex.mp.model import Model
        >>> mdl = Model()

        Create index-set:

        >>> nodes = IndexSet1D(['A', 'B', 'C'], name='node')

        Add variables:

        >>> from opti_extensions.docplex import add_variables
        >>> node_select = add_variables(mdl, nodes, 'B', name='node-select')

        Lookup for keys:

        >>> node_select.lookup('A')
        docplex.mp.Var(type=B,name='node-select_A')

        >>> node_select.lookup('Z')
        0
        """
        return super().get(key, 0)

    def sum(self) -> LinearExpr:
        """Sum all variables in an expression.

        Returns
        -------
        docplex.mp.linear.LinearExpr

        Examples
        --------
        Create DOcplex model:

        >>> from docplex.mp.model import Model
        >>> mdl = Model()

        Create index-set:

        >>> nodes = IndexSet1D(['A', 'B', 'C'], name='node')

        Add variables:

        >>> from opti_extensions.docplex import add_variables
        >>> node_select = add_variables(mdl, nodes, 'B', name='node-select')

        Sum all variables:

        >>> node_select.sum()
        docplex.mp.LinearExpr(node-select_A+node-select_B+node-select_C)
        """
        return self.model.sum_vars_all_different(self.values())

    def sum_squares(self) -> QuadExpr:
        """Sum squares of all variables in an expression.

        Returns
        -------
        docplex.mp.quad.QuadExpr

        Examples
        --------
        Create DOcplex model:

        >>> from docplex.mp.model import Model
        >>> mdl = Model()

        Create index-set:

        >>> nodes = IndexSet1D(['A', 'B', 'C'], name='node')

        Add variables:

        >>> from opti_extensions.docplex import add_variables
        >>> node_select = add_variables(mdl, nodes, 'B', name='node-select')

        Sum squares of all variables:

        >>> node_select.sum_squares()
        docplex.mp.quad.QuadExpr(node-select_A^2+node-select_B^2+node-select_C^2)
        """
        return self.model.sumsq(self.values())

    def dot(self, paramdict: ParamDict1D[Elem1DT, ParamT]) -> LinearExpr | ZeroExpr:
        """Sum the products of variables with corresponding coef from ParamDict1D, in an expression.

        Assumes the coef to be zero if not found in the ParamDict1D.

        Equivalent to::

          model.sum(paramdict.get(k, 0) * v for k, v in vardict.items())

        Parameters
        ----------
        paramdict : ParamDict1D
            ParamDict1D to be used for dot product.

        Returns
        -------
        docplex.mp.linear.LinearExpr or docplex.mp.linear.ZeroExpr

        Raises
        ------
        TypeError
            If the paramdict is not as instance of ParamDict1D.

        Notes
        -----
        This method is equivalent to using the matrix multiplication operator ``@``.

        Both ``ParamDict1D @ VarDict1D`` and ``VarDict1D @ ParamDict1D`` will also produce the same
        result.

        Examples
        --------
        Create DOcplex model:

        >>> from docplex.mp.model import Model
        >>> mdl = Model()

        Create index-set:

        >>> nodes = IndexSet1D(['A', 'B', 'C'], name='node')

        Add variables:

        >>> from opti_extensions.docplex import add_variables
        >>> node_select = add_variables(mdl, nodes, 'B', name='node-select')

        Define parameter:

        >>> from opti_extensions import ParamDict1D
        >>> fixed_cost = ParamDict1D({'A': 100, 'B': 200})

        Compute dot product (three alternative ways):

        >>> node_select.dot(fixed_cost)
        docplex.mp.LinearExpr(100node-select_A+200node-select_B)

        >>> fixed_cost @ node_select
        docplex.mp.LinearExpr(100node-select_A+200node-select_B)

        >>> node_select @ fixed_cost
        docplex.mp.LinearExpr(100node-select_A+200node-select_B)
        """
        _validate_for_dot_1d(paramdict)
        return self.model.sum(paramdict.get(k, 0) * v for k, v in self.items())

    __matmul__ = dot

    __rmatmul__ = dot


class VarDictND(VarDictCore[ElemNDT, VarT], DictNDMixin[ElemNDT, VarT]):
    """Custom subclasses of `dict` to define DOcplex model variables with N-dim tuple keys.

    Parameters
    ----------
    docpx_var_dict : dict
        Dictionary of variable objects from docplex.
    indexset : IndexSetND
        Keys of dictionary encapsulated in IndexSetND.
    model : docplex.mp.model.Model
        DOcplex model associated with the variable objects.
    vartype : docplex.mp.vartype.VarType
        DOcplex VarType corresponding to the variable objects.
    value_name : str, optional
        Name to refer to variables - not used internally, and solely for user reference.
    """

    # Private attributes
    # ------------------
    # _indexset : IndexSetND
    #     Index-set of keys.

    __slots__ = ('_key_names', '_value_name')

    def __init__(
        self,
        docpx_var_dict: dict[ElemNDT, VarT],
        indexset: IndexSetND[ElemNDT],
        /,
        *,
        model: Model,
        vartype: VarType,
        value_name: str | None = None,
    ) -> None:
        self.key_names = indexset.names
        self.value_name = value_name

        super().__init__(docpx_var_dict, indexset=indexset, model=model, vartype=vartype)

    def __new__(
        cls,
        docpx_var_dict: dict[ElemNDT, VarT],
        indexset: IndexSetND[ElemNDT],
        /,
        *,
        model: Model,
        vartype: VarType,
        value_name: str | None = None,
    ) -> VarDictND[ElemNDT, VarT]:
        raise TypeError(
            'This class is not meant to be instantiated; VarDictND is built through the '
            'opti_extensions.docplex.add_variables function'
        )

    @classmethod
    def _create(
        cls,
        docpx_var_dict: dict[ElemNDT, VarT],
        indexset: IndexSetND[ElemNDT],
        /,
        *,
        model: Model,
        vartype: VarType,
        value_name: str | None = None,
    ) -> VarDictND[ElemNDT, VarT]:
        # Private method to construct VarDictND
        instance = super().__new__(cls)
        cls.__init__(
            instance, docpx_var_dict, indexset, model=model, vartype=vartype, value_name=value_name
        )
        return instance

    def __repr__(self) -> str:
        # Printable string representation.
        return f'{self._get_repr_header()}\n{super().__repr__()}'

    def _repr_pretty_(self, p, cycle: bool) -> None:  # type: ignore[no-untyped-def]
        # Pretty repr for IPython.
        # https://ipython.readthedocs.io/en/stable/api/generated/IPython.lib.pretty.html#extending
        # Since IPython is not typed, we'll add a type ignore comment here
        # The annotation for arg `p` is `IPython.lib.pretty.RepresentationPrinter`
        p.text(f'{self._get_repr_header()}\n')
        p.pretty(dict(self))

    def lookup(self, *key: Any) -> VarT | Literal[0]:
        """Get the variable for the specified key, or zero if it is not found.

        Parameters
        ----------
        *key : key

        Returns
        -------
        docplex.mp.dvar.Var or ``0``

        Examples
        --------
        Create DOcplex model:

        >>> from docplex.mp.model import Model
        >>> mdl = Model()

        Create index-set:

        >>> arcs = IndexSetND([('A', 'B'), ('B', 'C'), ('C', 'B')], names=['ori', 'des'])

        Add variables:

        >>> from opti_extensions.docplex import add_variables
        >>> arc_flow = add_variables(mdl, arcs, 'C', ub=10, name='arc-flow')

        Lookup with keys:

        >>> arc_flow.lookup('A', 'B')
        docplex.mp.Var(type=C,name='arc-flow_A_B',ub=10)

        >>> arc_flow.lookup('X', 'Y')
        0
        """
        if any((isinstance(k, Iterable) and not isinstance(k, str)) for k in key):
            raise TypeError('lookup key must be scalars (no iterables except string)')
        if len(key) != self._indexset._tuplelen:
            raise ValueError('lookup key length must be the same as that of N-dim tuple keys')
        return super().get(cast('ElemNDT', key), 0)

    def sum(self, *pattern: Any) -> LinearExpr | ZeroExpr:
        """Sum all variables, or a subset based on wildcard pattern, in an expression.

        Parameters
        ----------
        *pattern : Any, optional
            For subsets, the pattern requires one value for each dimension of the N-dim tuple key.
            The single-character string ``'*'`` (asterisk) can be used as a wildcard to represent
            all possible values for a dimension.

        Returns
        -------
        docplex.mp.linear.LinearExpr or docplex.mp.linear.ZeroExpr

        Raises
        ------
        TypeError
            If the pattern includes non-scalar(s).
        ValueError
            If the pattern is not the same as the length of N-dim tuple keys.
        ValueError
            If the pattern has no wildcard or all wildcards.

        Examples
        --------
        Create DOcplex model:

        >>> from docplex.mp.model import Model
        >>> mdl = Model()

        Create index-set:

        >>> arcs = IndexSetND([('A', 'B'), ('B', 'C'), ('C', 'B')], names=['ori', 'des'])

        Add variables:

        >>> from opti_extensions.docplex import add_variables
        >>> arc_flow = add_variables(mdl, arcs, 'C', ub=10, name='arc-flow')

        Sum all variables:

        >>> arc_flow.sum()
        docplex.mp.LinearExpr(arc-flow_A_B+arc-flow_B_C+arc-flow_C_B)

        Sum subset of variables having ``'B'`` at the second dimension index:

        >>> arc_flow.sum('*', 'B')
        docplex.mp.LinearExpr(arc-flow_A_B+arc-flow_C_B)

        Sum subset of variables having ``'Z'`` at the first dimension index:

        >>> arc_flow.sum('Z', '*')
        docplex.mp.ZeroExpr()
        """
        if pattern:
            return self.model.sum_vars_all_different(self.subset_values(*pattern))
        return self.model.sum_vars_all_different(self.values())

    def sum_squares(self, *pattern: Any) -> QuadExpr | ZeroExpr:
        """Sum squares of all variables, or a subset based on wildcard pattern, in an expression.

        Parameters
        ----------
        *pattern : Any, optional
            For subsets, the pattern requires one value for each dimension of the N-dim tuple key.
            The single-character string ``'*'`` (asterisk) can be used as a wildcard to represent
            all possible values for a dimension.

        Returns
        -------
        docplex.mp.quad.QuadExpr or docplex.mp.linear.ZeroExpr

        Raises
        ------
        TypeError
            If the pattern includes non-scalar(s).
        ValueError
            If the pattern is not the same as the length of N-dim tuple keys.
        ValueError
            If the pattern has no wildcard or all wildcards.

        Examples
        --------
        Create DOcplex model:

        >>> from docplex.mp.model import Model
        >>> mdl = Model()

        Create index-set:

        >>> arcs = IndexSetND([('A', 'B'), ('B', 'C'), ('C', 'B')], names=['ori', 'des'])

        Add variables:

        >>> from opti_extensions.docplex import add_variables
        >>> arc_flow = add_variables(mdl, arcs, 'C', ub=10, name='arc-flow')

        Sum squares of all variables:

        >>> arc_flow.sum_squares()
        docplex.mp.quad.QuadExpr(arc-flow_A_B^2+arc-flow_B_C^2+arc-flow_C_B^2)

        Sum squares of subset of variables having ``'B'`` at the second dimension index:

        >>> arc_flow.sum_squares('*', 'B')
        docplex.mp.quad.QuadExpr(arc-flow_A_B^2+arc-flow_C_B^2)

        Sum squares of subset of variables having ``'Z'`` at the first dimension index:

        >>> arc_flow.sum_squares('Z', '*')
        docplex.mp.ZeroExpr()
        """
        if pattern:
            return self.model.sumsq(self.subset_values(*pattern))
        return self.model.sumsq(self.values())

    def dot(self, paramdict: ParamDictND[ElemNDT, ParamT]) -> LinearExpr | ZeroExpr:
        """Sum the products of variables with corresponding coef from ParamDictND, in an expression.

        Assumes the coef to be zero if not found in the ParamDictND.

        Equivalent to::

          model.sum(paramdict.get(k, 0) * v for k, v in vardict.items())

        Parameters
        ----------
        paramdict : ParamDictND
            ParamDictND to be used for dot product; should have tuple keys of same length as
            VarDictND.

        Returns
        -------
        docplex.mp.linear.LinearExpr or docplex.mp.linear.ZeroExpr

        Raises
        ------
        TypeError
            If the paramdict is not as instance of ParamDictND.
        ValueError
            If the paramdict does not have tuple keys of same length as VarDictND.

        Notes
        -----
        This method is equivalent to using the matrix multiplication operator ``@``.

        Both ``ParamDictND @ VarDictND`` and ``VarDictND @ ParamDictND`` will also produce the same
        result.

        Examples
        --------
        Create DOcplex model:

        >>> from docplex.mp.model import Model
        >>> mdl = Model()

        Create index-set:

        >>> arcs = IndexSetND([('A', 'B'), ('B', 'C'), ('C', 'B')], names=['ori', 'des'])

        Add variables:

        >>> from opti_extensions.docplex import add_variables
        >>> arc_flow = add_variables(mdl, arcs, 'C', ub=10, name='arc-flow')

        Define parameter:

        >>> from opti_extensions import ParamDictND
        >>> cost = ParamDictND({('A', 'B'): 10, ('B', 'C'): 20})

        Compute dot product (three alternative ways):

        >>> arc_flow.dot(cost)
        docplex.mp.LinearExpr(10arc-flow_A_B+20arc-flow_B_C)

        >>> cost @ arc_flow
        docplex.mp.LinearExpr(10arc-flow_A_B+20arc-flow_B_C)

        >>> arc_flow @ cost
        docplex.mp.LinearExpr(10arc-flow_A_B+20arc-flow_B_C)
        """
        _validate_for_dot_Nd(self._indexset, paramdict)
        return self.model.sum(paramdict.get(k, 0) * v for k, v in self.items())

    __matmul__ = dot

    __rmatmul__ = dot



# Copyright 2025 Samarth Mistry
# This file is part of the `opti-extensions` package, which is released under
# the Apache Licence, Version 2.0 (http://www.apache.org/licenses/LICENSE-2.0).

"""Functions to add variable(s) to an xpress problem."""

from __future__ import annotations

from collections.abc import Mapping
from typing import Literal, overload

import xpress as xp

from .._index_sets import Elem1DT, ElemNDT, IndexSet1D, IndexSetND
from .._param_dicts import ParamDict1D, ParamDictND, ParamT
from ._var_dicts import VarDict1D, VarDictND


def _paramdictNd_as_attr(
    indexset: IndexSetND[ElemNDT],
    attr: ParamDictND[ElemNDT, ParamT],
    attr_type: Literal['lb', 'ub', 'threshold'],
) -> ParamDictND[ElemNDT, ParamT]:
    """Use ParamDictND as xpress variable attribute.

    Parameters
    ----------
    indexset : IndexSetND
        Index-set for defining the variables.
    attr : ParamDictND
        ParamDict to be used as the attribute.
    attr_type : str
        Attribute type, ``'lb'`` or ``'ub'`` or ``'threshold'``.

    Returns
    -------
    ParamDictND
        ParamDictND that can be used as xpress variable attribute for the `problem.addVariables`
        method.

    Raises
    ------
    ValueError
        ParamDict keys and indexset have tuple elements of different lengths.
    """
    if (not attr) or (attr and attr._indexset._tuplelen == indexset._tuplelen):
        return attr
    else:
        raise ValueError(f'{attr_type} keys and indexset have tuple elements of different lengths')


def _preprocess_attr(
    indexset: IndexSet1D[Elem1DT] | IndexSetND[ElemNDT],
    attr: int
    | float
    | ParamDict1D[Elem1DT, ParamT]
    | ParamDictND[ElemNDT, ParamT]
    | Mapping[Elem1DT, ParamT]
    | Mapping[ElemNDT, ParamT],
    attr_type: Literal['lb', 'ub', 'threshold'],
) -> (
    int
    | float
    | ParamDict1D[Elem1DT, ParamT]
    | ParamDictND[ElemNDT, ParamT]
    | Mapping[Elem1DT, ParamT]
    | Mapping[ElemNDT, ParamT]
):
    """Preprocess xpress variable attribute.

    Parameters
    ----------
    indexset : IndexSet1D or IndexSetND
        Index-set for defining the variables.
    attr : Mapping or ParamDict
        Variable attribute.
    attr_type : str
        Attribute type, ``'lb'`` or ``'ub'`` or ``'threshold'``.

    Returns
    -------
    Mapping or ParamDict
        Xpress variable attribute for the `problem.addVariables` method.

    Raises
    ------
    ValueError
        ParamDict keys and indexset have tuple elements of different lengths.
    TypeError
        If paramdict is ParamDictND and indexset is IndexSet1D, or vice versa.
    """
    # Validate if ParamDict is used as variable attribute
    if isinstance(attr, ParamDict1D):
        if isinstance(indexset, IndexSet1D):
            return attr
        else:
            raise TypeError(f'`{attr_type}` should be ParamDictND when indexset is IndexSetND')
    elif isinstance(attr, ParamDictND):
        if isinstance(indexset, IndexSetND):
            return _paramdictNd_as_attr(indexset, attr, attr_type)
        else:
            raise TypeError(f'`{attr_type}` should be ParamDict1D when indexset is IndexSet1D')
    # Let xpress handle everything else, so return as is
    else:
        return attr


@overload
def addVariables(  # numpydoc ignore=GL08
    problem: xp.problem,
    indexset: IndexSet1D[Elem1DT],
    name: str | None = ...,
    lb: int | float | ParamDict1D[Elem1DT, ParamT] | Mapping[Elem1DT, int | float] = ...,
    ub: int | float | ParamDict1D[Elem1DT, ParamT] | Mapping[Elem1DT, int | float] = ...,
    threshold: int | float | ParamDict1D[Elem1DT, ParamT] | Mapping[Elem1DT, int | float] = ...,
    vartype: int = ...,
) -> VarDict1D[Elem1DT, xp.var]: ...


@overload
def addVariables(  # numpydoc ignore=GL08
    problem: xp.problem,
    indexset: IndexSetND[ElemNDT],
    name: str | None = ...,
    lb: int | float | ParamDictND[ElemNDT, ParamT] | Mapping[ElemNDT, int | float] = ...,
    ub: int | float | ParamDictND[ElemNDT, ParamT] | Mapping[ElemNDT, int | float] = ...,
    threshold: int | float | ParamDictND[ElemNDT, ParamT] | Mapping[ElemNDT, int | float] = ...,
    vartype: int = ...,
) -> VarDictND[ElemNDT, xp.var]: ...


def addVariables(
    problem: xp.problem,
    indexset: IndexSet1D[Elem1DT] | IndexSetND[ElemNDT],
    name: str | None = 'x',
    lb: int
    | float
    | ParamDict1D[Elem1DT, ParamT]
    | ParamDictND[ElemNDT, ParamT]
    | Mapping[Elem1DT, int | float]
    | Mapping[ElemNDT, int | float] = 0,
    ub: int
    | float
    | ParamDict1D[Elem1DT, ParamT]
    | ParamDictND[ElemNDT, ParamT]
    | Mapping[Elem1DT, int | float]
    | Mapping[ElemNDT, int | float] = xp.infinity,
    threshold: int
    | float
    | ParamDict1D[Elem1DT, ParamT]
    | ParamDictND[ElemNDT, ParamT]
    | Mapping[Elem1DT, int | float]
    | Mapping[ElemNDT, int | float] = 1,
    vartype: int = xp.continuous,
) -> VarDict1D[Elem1DT, xp.var] | VarDictND[ElemNDT, xp.var]:
    """Create and add multiple variables (corresponding to an index-set) to the xpress problem.

    This function extends xpress's `problem.addVariables` method by wrapping it and returning
    VarDict1D/VarDictND data structures - that can efficiently sum subsets of variables with
    a user-friendly syntax based on wildcard patterns.

    Parameters
    ----------
    problem : xpress.problem
        Xpress problem.
    indexset : IndexSet1D or IndexSetND
        Index-set for defining the variables. The ``name``/``names`` attribute of
        IndexSet1D/IndexSetND will be set as the ``key_name``/``key_names`` attribute of
        VarDict1D/VarDictND.
    name : str, optional
        Variable name, by default ``'x'``. It will also be used as the ``value_name`` attribute of
        VarDict1D/VarDictND.
    lb : int or float or dict or ParamDict, optional
        Lower bound, in one of the following forms:

        * A number - if all variables share the same lower bound.
        * A dict/ParamDict - with keys following the same structute as the index-set elements and
          values representing the lower bound; will fallback to the default for index-set elements
          not found in dict/ParamDict keys.

        Default is ``0.0``.
    ub : int or float or dict or ParamDict, optional
        Upper bound, in one of the following forms:

        * A number - if all variables share the same upper bound.
        * A dict/ParamDict - with keys following the same structute as the index-set elements and
          values representing the upper bound; will fallback to the default for index-set elements
          not found in dict/ParamDict keys.

        Default is ``1.0`` for binary variable type and ``infinity`` for other types.
    threshold : int or float or dict or ParamDict, optional
        Threshold, in one of the following forms:

        * A number - if all variables share the same objective coefficient.
        * A dict/ParamDict - with keys following the same structute as the index-set elements and
          values representing the objective coefficient; will fallback to the default for index-set
          elements not found in dict/ParamDict keys.

        Only applies to semi-continuous, semi-integer, and partially integer variables; it must be
        between its lower and its upper bound. Default is ``1``.
    vartype : int
        Variable type, in one of the following forms:

        * ``xpress.continuous``
        * ``xpress.binary``
        * ``xpress.integer``
        * ``xpress.semicontinuous``
        * ``xpress.semiinteger``
        * ``xpress.partiallyinteger``

        Default is ``xpress.continuous``.

    Returns
    -------
    VarDict1D or VarDictND

    Raises
    ------
    ValueError
        If the index-set is empty.

    Examples
    --------
    Create xpress problem:

    >>> import xpress as xp
    >>> import warnings
    >>> warnings.filterwarnings('ignore', category=xp.LicenseWarning)

    >>> prob = xp.problem()

    Create index-sets:

    >>> nodes = IndexSet1D(['A', 'B', 'C'], name='node')
    >>> arcs = IndexSetND([('A', 'B'), ('B', 'C'), ('C', 'B')], names=['ori', 'des'])

    Add variables:

    >>> node_select = addVariables(prob, nodes, name='select', vartype=xp.binary)
    >>> node_select
    VarDict1D: node -> select
    {'A': select(A), 'B': select(B), 'C': select(C)}

    >>> arc_flow = addVariables(prob, arcs, name='flow', ub=10, vartype=xp.continuous)
    >>> arc_flow
    VarDictND: (ori, des) -> flow
    {('A', 'B'): flow(('A', 'B')),
     ('B', 'C'): flow(('B', 'C')),
     ('C', 'B'): flow(('C', 'B'))}
    """
    if not isinstance(problem, xp.problem):
        raise TypeError('`problem` should be xp.problem')
    if not isinstance(indexset, IndexSet1D | IndexSetND):
        raise TypeError('`indexset` should be either IndexSet1D or IndexSetND')
    if not indexset:
        raise ValueError(f'{indexset.__class__.__name__} is empty')

    scalar_lb = isinstance(lb, int | float)
    scalar_ub = isinstance(ub, int | float)
    scalar_th = isinstance(threshold, int | float)
    if scalar_lb and scalar_ub and scalar_th:
        xp_var_dict = problem.addVariables(
            indexset, name=name, lb=lb, ub=ub, threshold=threshold, vartype=vartype
        )
    else:
        lb = _preprocess_attr(indexset, lb, 'lb')  # type: ignore[arg-type]
        ub = _preprocess_attr(indexset, ub, 'ub')  # type: ignore[arg-type]
        threshold = _preprocess_attr(indexset, threshold, 'threshold')  # type: ignore[arg-type]
        # Using type ignore as mypy cannot deduce typevars.
        xp_var_dict = {
            elem: problem.addVariable(
                name=f'{name}({elem})',
                lb=lb if scalar_lb else lb.get(elem, 0),  # type: ignore
                ub=ub if scalar_ub else ub.get(elem, xp.infinity),  # type: ignore
                threshold=threshold if scalar_th else threshold.get(elem, 1),  # type: ignore
                # Using type ignore as mypy cannot deduce that lb / ub / threshold would reduce to
                # paramdicts or mappings. Moreover, mypy also complains about the type of defaults
                # provided with respect to the typevars in class def.
                vartype=vartype,
            )
            for elem in indexset
        }

    value_name = name if isinstance(name, str) else None

    if isinstance(indexset, IndexSet1D):
        return VarDict1D._create(
            xp_var_dict, indexset, problem=problem, vartype=vartype, value_name=value_name
        )
    return VarDictND._create(
        xp_var_dict, indexset, problem=problem, vartype=vartype, value_name=value_name
    )



# Copyright 2025 Samarth Mistry
# This file is part of the `opti-extensions` package, which is released under
# the Apache Licence, Version 2.0 (http://www.apache.org/licenses/LICENSE-2.0).

"""Functions to add variable(s) to a gurobipy model."""

from __future__ import annotations

from collections.abc import Mapping, Sequence
from typing import Literal, overload

from gurobipy import GRB, Model, Var

from .._index_sets import Elem1DT, ElemNDT, IndexSet1D, IndexSetND
from .._param_dicts import ParamDict1D, ParamDictND, ParamT
from ._var_dicts import VarDict1D, VarDictND


def _paramdictNd_as_attr(
    indexset: IndexSetND[ElemNDT],
    attr: ParamDictND[ElemNDT, ParamT],
    attr_type: Literal['lb', 'ub', 'obj'],
) -> ParamDictND[ElemNDT, ParamT]:
    """Use ParamDictND as gurobipy variable attribute.

    Parameters
    ----------
    indexset : IndexSetND
        Index-set for defining the variables.
    attr : ParamDictND
        ParamDict to be used as the attribute.
    attr_type : str
        Attribute type, ``'lb'`` or ``'ub'`` or ``'obj'``.

    Returns
    -------
    ParamDictND
        ParamDictND that can be used as gurobipy variable attribute for the `Model.addVars` method.

    Raises
    ------
    ValueError
        ParamDict keys and indexset have tuple elements of different lengths.
    """
    if (not attr) or (attr and attr._indexset._tuplelen == indexset._tuplelen):
        return attr
    else:
        raise ValueError(f'{attr_type} keys and indexset have tuple elements of different lengths')


def _preprocess_attr(
    indexset: IndexSet1D[Elem1DT] | IndexSetND[ElemNDT],
    attr: int
    | float
    | Sequence[int | float]
    | ParamDict1D[Elem1DT, ParamT]
    | ParamDictND[ElemNDT, ParamT]
    | Mapping[Elem1DT, int | float]
    | Mapping[ElemNDT, int | float],
    attr_type: Literal['lb', 'ub', 'obj'],
) -> (
    int
    | float
    | Sequence[int | float]
    | ParamDict1D[Elem1DT, ParamT]
    | ParamDictND[ElemNDT, ParamT]
    | Mapping[Elem1DT, int | float]
    | Mapping[ElemNDT, int | float]
):
    """Preprocess gurobipy variable attribute.

    Parameters
    ----------
    indexset : IndexSet1D or IndexSetND
        Index-set for defining the variables.
    attr : int or float or Sequence or Mapping or ParamDict
        Variable attribute.
    attr_type : str
        Attribute type, ``'lb'`` or ``'ub'`` or ``'obj'``.

    Returns
    -------
    int or float or Sequence or Mapping or ParamDict
        The gurobipy variable attribute for the `Model.addVars` method.

    Raises
    ------
    ValueError
        ParamDict keys and indexset have tuple elements of different lengths.
    TypeError
        If paramdict is ParamDictND and indexset is IndexSet1D, or vice versa.
    """
    # Validate if ParamDict is used as variable attribute
    if isinstance(attr, ParamDict1D):
        if isinstance(indexset, IndexSet1D):
            return attr
        else:
            raise TypeError(f'`{attr_type}` should be ParamDictND when indexset is IndexSetND')
    elif isinstance(attr, ParamDictND):
        if isinstance(indexset, IndexSetND):
            return _paramdictNd_as_attr(indexset, attr, attr_type)
        else:
            raise TypeError(f'`{attr_type}` should be ParamDict1D when indexset is IndexSet1D')
    # Let gurobipy handle everything else, so return as is
    else:
        return attr


@overload
def addVars(  # numpydoc ignore=GL08
    model: Model,
    indexset: IndexSet1D[Elem1DT],
    lb: int
    | float
    | Sequence[int | float]
    | ParamDict1D[Elem1DT, ParamT]
    | Mapping[Elem1DT, int | float] = ...,
    ub: int
    | float
    | Sequence[int | float]
    | ParamDict1D[Elem1DT, ParamT]
    | Mapping[Elem1DT, int | float] = ...,
    obj: int
    | float
    | Sequence[int | float]
    | ParamDict1D[Elem1DT, ParamT]
    | Mapping[Elem1DT, int | float] = ...,
    vtype: str = ...,
    name: str = ...,
) -> VarDict1D[Elem1DT, Var]: ...


@overload
def addVars(  # numpydoc ignore=GL08
    model: Model,
    indexset: IndexSetND[ElemNDT],
    lb: int
    | float
    | Sequence[int | float]
    | ParamDictND[ElemNDT, ParamT]
    | Mapping[ElemNDT, int | float] = ...,
    ub: int
    | float
    | Sequence[int | float]
    | ParamDictND[ElemNDT, ParamT]
    | Mapping[ElemNDT, int | float] = ...,
    obj: int
    | float
    | Sequence[int | float]
    | ParamDictND[ElemNDT, ParamT]
    | Mapping[ElemNDT, int | float] = ...,
    vtype: str = ...,
    name: str = ...,
) -> VarDictND[ElemNDT, Var]: ...


def addVars(
    model: Model,
    indexset: IndexSet1D[Elem1DT] | IndexSetND[ElemNDT],
    lb: int
    | float
    | Sequence[int | float]
    | Mapping[Elem1DT, int | float]
    | Mapping[ElemNDT, int | float]
    | ParamDict1D[Elem1DT, ParamT]
    | ParamDictND[ElemNDT, ParamT] = 0.0,
    ub: int
    | float
    | Sequence[int | float]
    | ParamDict1D[Elem1DT, ParamT]
    | ParamDictND[ElemNDT, ParamT]
    | Mapping[Elem1DT, int | float]
    | Mapping[ElemNDT, int | float] = float('inf'),
    obj: int
    | float
    | Sequence[int | float]
    | ParamDict1D[Elem1DT, ParamT]
    | ParamDictND[ElemNDT, ParamT]
    | Mapping[Elem1DT, int | float]
    | Mapping[ElemNDT, int | float] = 0.0,
    vtype: str = GRB.CONTINUOUS,
    name: str = '',
) -> VarDict1D[Elem1DT, Var] | VarDictND[ElemNDT, Var]:
    """Create and add multiple variables (corresponding to an index-set) to the gurobipy model.

    This function extends gurobipy's `Model.addVars` method by wrapping it and returning
    VarDict1D/VarDictND data structures - that can efficiently sum subsets of variables with
    a user-friendly syntax based on wildcard patterns.

    Parameters
    ----------
    model : gurobipy.Model
        The gurobipy model.
    indexset : IndexSet1D or IndexSetND
        Index-set for defining the variables. The ``name``/``names`` attribute of
        IndexSet1D/IndexSetND will be set as the ``key_name``/``key_names`` attribute of
        VarDict1D/VarDictND.
    lb : int or float or sequence or dict or ParamDict, optional
        Lower bound, in one of the following forms:

        * A number - if all variables share the same lower bound.
        * A sequence of numbers - one for each variable.
        * A dict/ParamDict - with keys following the same structute as the index-set elements and
          values representing the lower bound; will fallback to the default for index-set elements
          not found in dict/ParamDict keys.

        Default is ``0.0``.
    ub : int or float or sequence or dict or ParamDict, optional
        Upper bound, in one of the following forms:

        * A number - if all variables share the same upper bound.
        * A sequence of numbers - one for each variable.
        * A dict/ParamDict - with keys following the same structute as the index-set elements and
          values representing the upper bound; will fallback to the default for index-set elements
          not found in dict/ParamDict keys.

        Default is ``1.0`` for binary variable type and ``infinity`` for other types.
    obj : int or float or sequence or dict or ParamDict, optional
        Objective coefficient, in one of the following forms:

        * A number - if all variables share the same objective coefficient.
        * A sequence of numbers - one for each variable.
        * A dict/ParamDict - with keys following the same structute as the index-set elements and
          values representing the objective coefficient; will fallback to the default for index-set
          elements not found in dict/ParamDict keys.

        Default is ``0.0``.
    vtype : str, optional
        Variable type, in one of the following forms:

        * ``GRB.CONTINUOUS``
        * ``GRB.BINARY``
        * ``GRB.INTEGER``
        * ``GRB.SEMICONT``
        * ``GRB.SEMIINT``

        Default is ``GRB.CONTINUOUS``.
    name : str, optional
        Variable name, by default empty string. It will also be used as the ``value_name`` attribute
        of VarDict1D/VarDictND.

    Returns
    -------
    VarDict1D or VarDictND

    Raises
    ------
    ValueError
        If the index-set is empty.

    Examples
    --------
    Create gurobipy model:

    >>> from gurobipy import Model
    >>> mdl = Model()

    Create index-sets:

    >>> nodes = IndexSet1D(['A', 'B', 'C'], name='node')
    >>> arcs = IndexSetND([('A', 'B'), ('B', 'C'), ('C', 'B')], names=['ori', 'des'])

    Add variables:

    >>> node_select = addVars(mdl, nodes, vtype=GRB.BINARY, name='select')
    >>> mdl.update()
    >>> node_select
    VarDict1D: node -> select
    {'A': <gurobi.Var select[A]>, 'B': <gurobi.Var select[B]>, 'C': <gurobi.Var select[C]>}

    >>> arc_flow = addVars(mdl, arcs, ub=10, vtype=GRB.CONTINUOUS, name='flow')
    >>> mdl.update()
    >>> arc_flow
    VarDictND: (ori, des) -> flow
    {('A', 'B'): <gurobi.Var flow[A,B]>,
     ('B', 'C'): <gurobi.Var flow[B,C]>,
     ('C', 'B'): <gurobi.Var flow[C,B]>}
    """
    if not isinstance(model, Model):
        raise TypeError('`model` should be gurobipy.Model')
    if not isinstance(indexset, IndexSet1D | IndexSetND):
        raise TypeError('`indexset` should be either IndexSet1D or IndexSetND')
    if not indexset:
        raise ValueError(f'{indexset.__class__.__name__} is empty')

    lb = _preprocess_attr(indexset, lb, 'lb')
    ub = _preprocess_attr(indexset, ub, 'ub')
    obj = _preprocess_attr(indexset, obj, 'obj')

    gp_var_dict = model.addVars(indexset, lb=lb, ub=ub, obj=obj, vtype=vtype, name=name)  # type: ignore[arg-type]
    # Since gurobipy stubs do not identify IndexSet and ParamDict data structures, we'll add a type
    # ignore comment here
    value_name = name if isinstance(name, str) else None

    if isinstance(indexset, IndexSet1D):
        return VarDict1D._create(
            gp_var_dict, indexset, model=model, vtype=vtype, value_name=value_name
        )
    return VarDictND._create(gp_var_dict, indexset, model=model, vtype=vtype, value_name=value_name)



# Copyright 2025 Samarth Mistry
# This file is part of the `opti-extensions` package, which is released under
# the Apache Licence, Version 2.0 (http://www.apache.org/licenses/LICENSE-2.0).

"""Tuning tool functionality."""

from io import TextIOBase
from operator import attrgetter
from pathlib import Path
from typing import Any, TextIO

from docplex.mp.model import Model
from docplex.mp.params.parameters import Parameter
from docplex.mp.utils import DOcplexException


def get_name(parameter: Any) -> str:
    """Get name of the parameter in the form a string.

    Parameters
    ----------
    parameter : DOcplex parameter or CPLEX legacy parameter object

    Returns
    -------
    str
        Parameter name.
    """
    # Cannot typecheck function arg since CPLEX legacy parameter class is not exposed to the public
    # API
    if isinstance(parameter, Parameter):
        name: str = parameter.qualified_name
        return name.split('.', maxsplit=1)[1]
    else:
        return str(parameter).split('.', maxsplit=1)[1]


def _tune(
    input: Model | list[str],
    /,
    *,
    log_output: bool | str | TextIO | None = True,
    overall_timelimit_sec: int | float | str | None = None,
    tuning_timelimit_sec: int | float | str | None = None,
    overall_timelimit_det: int | float | str | None = None,
    tuning_timelimit_det: int | float | str | None = None,
    display_level: int | None = None,
    measure: int | None = None,
    repeat: int | None = None,
    fixed_params_and_values: dict[str, int | float | bool | str] | None = None,
) -> dict[str, int | float]:
    """Run the CPLEX tuning tool for a DOcplex model or a batch of models (stored as files).

    Parameters
    ----------
    input : docplex.mp.model.Model or list[str]
        Either the DOcplex model or list of filepaths of the models to be tuned.
    log_output : bool or str or stream object, optional
        Log output switch, in one of the following forms:

        * ``True`` or ``'1'`` or ``'stdout'`` or ``'sys.stdout'``: Log is output to stdout.
        * ``'stderr'`` or ``'sys.stderr'``: Log is output to stderr.
        * ``False`` or ``'0'`` or ``None``: No log output.
        * File path (in form of str): Log is output to the file.
        * Stream object (a file-like object with a write method and a flush method): Log is output
          to the stream object.

        Default is ``True``.
    overall_timelimit_sec : int or float or str, optional
        Time limit for the tuning tool, in terms of seconds. Corresponds to the CPLEX parameter
        `timelimit`.
    tuning_timelimit_sec : int or float or str, optional
        Time limit for each test of the tuning tool, in terms of seconds. Corresponds to the CPLEX
        parameter `tune.timelimit`.
    overall_timelimit_det : int or float or str, optional
        Time limit for the tuning tool, in terms of deterministic ticks. Corresponds to the CPLEX
        parameter `dettimelimit`.
    tuning_timelimit_det : int or float or str, optional
        Time limit for each test of the tuning tool, in terms of deterministic ticks. Corresponds to
        the CPLEX parameter `tune.dettimelimit`.
    display_level : int, optional
        Level of information reported by the tuning tool as it works. Corresponds to the CPLEX
        parameter `tune.display`.
    measure : int, optional
        Measure for evaluating progress when a batch of models is being tuned. Only applicable when
        tuning a batch of models. Corresponds to the CPLEX parameter `tune.measure`.
    repeat : int, optional
        Number of times tuning is to be repeated on reordered versions of a given model. Only
        applicable when tuning a single model. Corresponds to the CPLEX parameter `tune.repeat`.
    fixed_params_and_values : dict, optional
        Set of parameters and their values that should be respected by the tuning tool, in form of a
        dict having parameter names (as str such as 'lpmethod', 'mip.limits.cutpasses', etc.) as
        keys and parameter values as values. Default is ``None``.

    Returns
    -------
    dict
        Set of performance-improving parameters and their values identified by tuning tool, in form
        of a dict having parameter names (as str such as 'lpmethod', 'mip.limits.cutpasses', etc.)
        as keys and parameter values as values.

    Raises
    ------
    ValueError
        If both overall_timelimit_sec and overall_timelimit_det are specified.
    ValueError
        If both tuning_timelimit_sec and tuning_timelimit_det are specified.
    ValueError
        If an invalid value is specified for any tuning parameter in function arguments:
        overall_timelimit_sec, tuning_timelimit_sec, overall_timelimit_det, tuning_timelimit_det,
        display_level, measure, repeat.
    ValueError
        If an invalid parameter or value is specified is fixed_params_and_values.
    """
    # Validation checks
    if isinstance(log_output, str) and log_output in ('1', '0'):
        log_output = bool(int(log_output))
    if not (
        log_output is None
        or isinstance(log_output, bool | str | TextIOBase)
        or (hasattr(log_output, 'write') and hasattr(log_output, 'flush'))
    ):
        raise TypeError('`log_output` should be a bool or str or stream object')
    if overall_timelimit_sec and overall_timelimit_det:
        raise ValueError(
            'Only one of `overall_timelimit_sec` and `overall_timelimit_det` can be set at a time'
        )
    if tuning_timelimit_sec and tuning_timelimit_det:
        raise ValueError(
            'Only one of `tuning_timelimit_sec` and `tuning_timelimit_det` can be set at a time'
        )
    if fixed_params_and_values is None:
        fixed_params_and_values = {}
    if not isinstance(fixed_params_and_values, dict):
        raise TypeError('`fixed_params_and_values` should be a dict')

    # Process input
    match input:
        case Model():
            batch_mode = False
            model = input
            # Register prior state so it can be restored post-tuning
            prior_log_output = model.log_output
            prior_params = {
                get_name(param): param.get()
                for param in model.parameters.generate_nondefault_params()
            }
            model.parameters.reset_all()
        case list():
            batch_mode = True
            model = Model()  # Dummy model to be used as proxy for accessing CPLEX env

    # Set params in the DOcplex model for the tuning tool
    tuning_args = {
        'overall_timelimit_sec': (overall_timelimit_sec, model.parameters.timelimit),
        'tuning_timelimit_sec': (tuning_timelimit_sec, model.parameters.tune.timelimit),
        'overall_timelimit_det': (overall_timelimit_det, model.parameters.dettimelimit),
        'tuning_timelimit_det': (tuning_timelimit_det, model.parameters.tune.dettimelimit),
        'display_level': (display_level, model.parameters.tune.display),
        'measure': (measure, model.parameters.tune.measure),
        'repeat': (repeat, model.parameters.tune.repeat),
    }
    tuning_params = {}
    for arg_name, (arg_val, param) in tuning_args.items():
        if arg_val is not None:
            try:
                param.set(arg_val)
            except (ValueError, DOcplexException):
                raise ValueError(f'Value {arg_val} for `{arg_name}` is invalid') from None
        tuning_params[get_name(param)] = param.get()

    # Sync DOcplex model params with the underlying CPLEX engine
    model.apply_parameters()

    # Access the underlying CPLEX engine through legacy API
    cplex = model.get_cplex()

    # Fixed params to be set directly in the CPLEX engine
    if fixed_params_and_values:
        param_set = cplex.create_parameter_set()
        for fixed_name, fixed_val in fixed_params_and_values.items():
            try:
                param = attrgetter(fixed_name)(model.parameters)
            except AttributeError:
                raise ValueError(
                    f'`{fixed_name}` in `fixed_parameters_and_values` is not a valid parameter'
                ) from None
            else:
                if param.accept_value(fixed_val):
                    param_set.add(attrgetter(fixed_name)(cplex.parameters), fixed_val)
                else:
                    # Provide a more informative error message since the one from CPLEX engine
                    # is ambiguous
                    raise ValueError(
                        f'Value {fixed_val} for `{fixed_name}` in `fixed_parameters_and_values` is '
                        'invalid'
                    )
    else:
        param_set = None

    if log_output:
        model.log_output = log_output
        stream = model.context.solver.log_output_as_stream

        if cplex.get_problem_name() != model.name:
            cplex.set_problem_name(model.name)

        # Log tuning tool params
        log_header = 'Tuning tool parameters:\n'
        for tuning_name, tuning_val in tuning_params.items():
            if not (tuning_name == 'tune.repeat' and batch_mode) and not (
                tuning_name == 'tune.measure' and not batch_mode
            ):
                log_header += f'   {tuning_name:<40} {tuning_val}\n'

        # Log fixed params
        log_header += '\nFixed parameters:\n'
        if fixed_params_and_values:
            for fixed_name, fixed_val in fixed_params_and_values.items():
                log_header += f'   {fixed_name:<40} {fixed_val}\n'
        else:
            log_header += '   None\n'

        stream.write(log_header)
        stream.flush()

    # Run the tuning tool
    if batch_mode:
        status = cplex.parameters.tune_problem_set(input, fixed_parameters_and_values=param_set)
    else:
        status = cplex.parameters.tune_problem(param_set)

    improving_params_and_values = {}
    # Exclude from coverage since there's no clear way to deterministically test the tuning tool
    # across multiple CPLEX versions. Also, it's not critical as it's only populating a dict.
    for param, param_value in cplex.parameters.get_changed():  # pragma: no cover
        param_name = get_name(param)
        if param_name not in fixed_params_and_values and param_name not in tuning_params:
            improving_params_and_values[param_name] = param_value

    if log_output:
        # Log status
        log_footer = f'Tuning tool status: {cplex.parameters.tuning_status[status]}\n\n'

        # Log improving params
        if improving_params_and_values:  # pragma: no cover
            log_footer += 'Tuned parameters:\n'
            for tuned_name, tuned_val in improving_params_and_values.items():
                log_footer += f'   {tuned_name:<40} {tuned_val}\n'
        else:
            log_footer += 'The tuning tool could not find a better set of parameters\n'

        stream.write(log_footer)
        stream.flush()
        try:
            stream.custom_close()
        except AttributeError:
            pass

    # Post-tuning cleanup
    if batch_mode:
        # Terminate the dummy model
        model.end()
    else:
        # Restore prior state
        model.log_output = prior_log_output
        for tuning_param_name in tuning_params:
            attrgetter(tuning_param_name)(model.parameters).reset()
        for prior_name, prior_val in prior_params.items():
            attrgetter(prior_name)(model.parameters).set(prior_val)
        model.apply_parameters()

    return improving_params_and_values


def tune(
    model: Model,
    *,
    log_output: bool | str | TextIO | None = True,
    overall_timelimit_sec: int | float | str | None = None,
    tuning_timelimit_sec: int | float | str | None = None,
    overall_timelimit_det: int | float | str | None = None,
    tuning_timelimit_det: int | float | str | None = None,
    display_level: int | None = None,
    repeat: int | None = None,
    fixed_params_and_values: dict[str, int | float | bool | str] | None = None,
) -> dict[str, int | float]:
    """Run tuning tests on a DOcplex model to search for performance-improving parameter settings.

    This function is a wrapper around CPLEX's tuning tool. Refer to the CPLEX user manual for more
    details about this functionality.

    Parameters
    ----------
    model : docplex.mp.model.Model
        DOcplex model.
    log_output : bool or str or stream object, optional
        Log output switch, in one of the following forms:

        * ``True`` or ``'1'`` or ``'stdout'`` or ``'sys.stdout'``: Log is output to stdout.
        * ``'stderr'`` or ``'sys.stderr'``: Log is output to stderr.
        * ``False`` or ``'0'`` or ``None``: No log output.
        * File path (in form of str): Log is output to the file.
        * Stream object (a file-like object with a write method and a flush method): Log is output
          to the stream object.

        Default is ``True``.
    overall_timelimit_sec : int or float or str, optional
        Time limit for the tuning tool, in terms of seconds. Corresponds to the CPLEX parameter
        `timelimit`.
    tuning_timelimit_sec : int or float or str, optional
        Time limit for each test of the tuning tool, in terms of seconds. Corresponds to the CPLEX
        parameter `tune.timelimit`.
    overall_timelimit_det : int or float or str, optional
        Time limit for the tuning tool, in terms of deterministic ticks. Corresponds to the CPLEX
        parameter `dettimelimit`.
    tuning_timelimit_det : int or float or str, optional
        Time limit for each test of the tuning tool, in terms of deterministic ticks. Corresponds to
        the CPLEX parameter `tune.dettimelimit`.
    display_level : int, optional
        Level of information reported by the tuning tool as it works. Corresponds to the CPLEX
        parameter `tune.display`.
    repeat : int, optional
        Number of times tuning is to be repeated on reordered versions of a given model. Corresponds
        to the CPLEX parameter `tune.repeat`.
    fixed_params_and_values : dict, optional
        Set of parameters and their values that should be respected by the tuning tool, in form of a
        dict having parameter names (as str such as 'lpmethod', 'mip.limits.cutpasses', etc.) as
        keys and parameter values as values. Default is ``None``.

    Returns
    -------
    dict
        Set of performance-improving parameters and their values identified by tuning tool, in form
        of a dict having parameter names (as str such as 'lpmethod', 'mip.limits.cutpasses', etc.)
        as keys and parameter values as values.

    Raises
    ------
    ValueError
        If both overall_timelimit_sec and overall_timelimit_det are specified.
    ValueError
        If both tuning_timelimit_sec and tuning_timelimit_det are specified.
    ValueError
        If an invalid value is specified for any tuning parameter in function arguments:
        overall_timelimit_sec, tuning_timelimit_sec, overall_timelimit_det, tuning_timelimit_det,
        display_level, repeat.
    ValueError
        If an invalid parameter or value is specified is fixed_params_and_values.

    See Also
    --------
    batch_tune : For tuning a batch of models (stored as files).

    Notes
    -----
    Any CPLEX parameters set for the model before passing it to this function will be disregarded.
    """
    if not isinstance(model, Model):
        raise TypeError('`model` should be docplex.mp.model.Model')

    return _tune(
        model,
        log_output=log_output,
        overall_timelimit_sec=overall_timelimit_sec,
        tuning_timelimit_sec=tuning_timelimit_sec,
        overall_timelimit_det=overall_timelimit_det,
        tuning_timelimit_det=tuning_timelimit_det,
        display_level=display_level,
        repeat=repeat,
        fixed_params_and_values=fixed_params_and_values,
    )


def batch_tune(
    *model_files: str,
    log_output: bool | str | TextIO | None = True,
    overall_timelimit_sec: int | float | str | None = None,
    tuning_timelimit_sec: int | float | str | None = None,
    overall_timelimit_det: int | float | str | None = None,
    tuning_timelimit_det: int | float | str | None = None,
    display_level: int | None = None,
    measure: int | None = None,
    fixed_params_and_values: dict[str, int | float | bool | str] | None = None,
) -> dict[str, int | float]:
    """Run tuning tests on a batch of models to search for performance-improving parameter settings.

    This function is a wrapper around CPLEX's tuning tool. Refer to the CPLEX user manual for more
    details about this functionality.

    Parameters
    ----------
    *model_files : str
        Path of file containing the model. Supported formats are ``.mps``, ``.lp``, and ``.sav``
        (compressed files with extensions ``.gz`` and ``.bz2`` are also supported).
    log_output : bool or str or stream object, optional
        Log output switch, in one of the following forms:

        * ``True`` or ``'1'`` or ``'stdout'`` or ``'sys.stdout'``: Log is output to stdout.
        * ``'stderr'`` or ``'sys.stderr'``: Log is output to stderr.
        * ``False`` or ``'0'`` or ``None``: No log output.
        * File path (in form of str): Log is output to the file.
        * Stream object (a file-like object with a write method and a flush method): Log is output
          to the stream object.

        Default is ``True``.
    overall_timelimit_sec : int or float or str, optional
        Time limit for the tuning tool, in terms of seconds. Corresponds to the CPLEX parameter
        `timelimit`.
    tuning_timelimit_sec : int or float or str, optional
        Time limit for each test of the tuning tool, in terms of seconds. Corresponds to the CPLEX
        parameter `tune.timelimit`.
    overall_timelimit_det : int or float or str, optional
        Time limit for the tuning tool, in terms of deterministic ticks. Corresponds to the CPLEX
        parameter `dettimelimit`.
    tuning_timelimit_det : int or float or str, optional
        Time limit for each test of the tuning tool, in terms of deterministic ticks. Corresponds to
        the CPLEX parameter `tune.dettimelimit`.
    display_level : int, optional
        Level of information reported by the tuning tool as it works. Corresponds to the CPLEX
        parameter `tune.display`.
    measure : int, optional
        Measure for evaluating progress. Corresponds to the CPLEX parameter `tune.measure`.
    fixed_params_and_values : dict, optional
        Set of parameters and their values that should be respected by the tuning tool, in form of a
        dict having parameter names (as str such as 'lpmethod', 'mip.limits.cutpasses', etc.) as
        keys and parameter values as values. Default is ``None``.

    Returns
    -------
    dict
        Set of performance-improving parameters and their values identified by tuning tool, in form
        of a dict having parameter names (as str such as 'lpmethod', 'mip.limits.cutpasses', etc.)
        as keys and parameter values as values.

    Raises
    ------
    FileNotFoundError
        If model file is not found.
    ValueError
        If both overall_timelimit_sec and overall_timelimit_det are specified.
    ValueError
        If both tuning_timelimit_sec and tuning_timelimit_det are specified.
    ValueError
        If an invalid value is specified for any tuning parameter in function arguments:
        overall_timelimit_sec, tuning_timelimit_sec, overall_timelimit_det, tuning_timelimit_det,
        display_level, measure.
    ValueError
        If an invalid parameter or value is specified is fixed_params_and_values.

    See Also
    --------
    tune : For tuning a DOcplex model.

    Notes
    -----
    Although the tuning tool can tune a single model through this function, it cannot access the
    CPLEX parameter `tune.repeat`. Use the ``opti_extensions.docplex.tune`` function instead to use
    this parameter.
    """
    for file in model_files:
        if isinstance(file, Model):
            raise TypeError(
                'batch_tune does not work with DOcplex models - use model files such as .mps, .lp, '
                'or .sav (compressed files with extensions .gz and .bz2 are also supported)'
            )
        if not isinstance(file, str):
            raise TypeError(f'File path `{file}` should be a string')
        if not Path(file).is_file():
            raise FileNotFoundError(f'File `{file}` not found')

    return _tune(
        list(model_files),
        log_output=log_output,
        overall_timelimit_sec=overall_timelimit_sec,
        tuning_timelimit_sec=tuning_timelimit_sec,
        overall_timelimit_det=overall_timelimit_det,
        tuning_timelimit_det=tuning_timelimit_det,
        display_level=display_level,
        measure=measure,
        fixed_params_and_values=fixed_params_and_values,
    )



# Copyright 2025 Samarth Mistry
# This file is part of the `opti-extensions` package, which is released under
# the Apache Licence, Version 2.0 (http://www.apache.org/licenses/LICENSE-2.0).

"""
The `opti-extensions` package.

`docplex`-specific functionality.
"""

# Check required dependencies
from importlib.util import find_spec as _find_spec

if _find_spec('docplex') is None:
    raise ImportError('Unable to import required dependency: docplex')

# Package functionality
from ._model_funcs import print_problem_stats, print_solution_quality_stats, runseeds, solve
from ._tuning_funcs import batch_tune, tune
from ._var_dicts import VarDict1D, VarDictND
from ._var_funcs import add_variable, add_variables

# Public API
__all__ = [
    'print_problem_stats',
    'print_solution_quality_stats',
    'runseeds',
    'solve',
    'batch_tune',
    'tune',
    'VarDict1D',
    'VarDictND',
    'add_variable',
    'add_variables',
]



# Copyright 2025 Samarth Mistry
# This file is part of the `opti-extensions` package, which is released under
# the Apache Licence, Version 2.0 (http://www.apache.org/licenses/LICENSE-2.0).

"""
The `opti-extensions` package.

A collection of custom data structures and user-friendly functions for mathematical optimization
modeling with DOcplex, gurobipy, and Xpress.
"""

__version__ = '1.2.0'

# Package functionality
from ._index_sets import IndexSet1D, IndexSetND
from ._pandas_accessors import DataFrameAccessor as _DataFrameAccessor
from ._pandas_accessors import IndexAccessor as _IndexAccessor
from ._pandas_accessors import SeriesAccessor as _SeriesAccessor
from ._param_dicts import ParamDict1D, ParamDictND

# Register custom accessors if pandas is available
try:
    import pandas as _pd

    _pd.api.extensions.register_series_accessor('opti')(_SeriesAccessor)
    _pd.api.extensions.register_dataframe_accessor('opti')(_DataFrameAccessor)
    _pd.api.extensions.register_index_accessor('opti')(_IndexAccessor)

except ImportError:
    pass

# Public API
__all__ = ['IndexSet1D', 'IndexSetND', 'ParamDict1D', 'ParamDictND']



# Copyright 2025 Samarth Mistry
# This file is part of the `opti-extensions` package, which is released under
# the Apache Licence, Version 2.0 (http://www.apache.org/licenses/LICENSE-2.0).

"""Accessors to cast pandas Series/DataFrame/Index into IndexSet and ParamDict data structures."""

from __future__ import annotations

from collections.abc import Iterable
from typing import TYPE_CHECKING, Any

from ._index_sets import IndexSet1D, IndexSetND
from ._param_dicts import ParamDict1D, ParamDictND

if TYPE_CHECKING:
    from pandas import DataFrame, Index, Series


def _check_empty(pd_obj: Series[Any] | DataFrame | Index[Any]) -> None:
    """Raise a ValueError if a pandas object is empty.

    Parameters
    ----------
    pd_obj : Series or DataFrame or Index
    """
    if pd_obj.empty:
        raise ValueError(f'{pd_obj.__class__.__name__} is empty')


def _check_duplicate_idx(pd_obj: Series[Any] | DataFrame) -> None:
    """Raise a ValueError if a pandas object has duplicate index-label(s).

    Parameters
    ----------
    pd_obj : Series or DataFrame
    """
    if pd_obj.index.size > len(set(pd_obj.index)):
        raise ValueError(f'{pd_obj.__class__.__name__} has duplicate index-label(s)')


class DataFrameAccessor:
    """Accessor to cast pandas DataFrame into IndexSet1D/IndexSetND and ParamDict1D/ParamDictND.

    Parameters
    ----------
    df : DataFrame
    """

    def __init__(self, df: DataFrame):
        _check_empty(df)
        self._df = df

    def to_indexset(self) -> IndexSet1D[Any] | IndexSetND[tuple[Any, ...]]:
        """Cast a DataFrame into an IndexSet1D/IndexSetND.

        Note: The `opti-extensions` package has to be imported first to use this method with
        `pandas`.

        Returns
        -------
        IndexSet1D or IndexSetND
            * A single-column DataFrame will be cast into an IndexSet1D consisting of its column
              values. The column name will be set as the `IndexSet1D.name` attribute (as str).
            * A multi-column DataFrame will be cast into an IndexSetND consisting of tuples, each
              containing row values. The column names will be set as the `IndexSetND.names`
              attribute (as list[str]).

        Raises
        ------
        ValueError
            If the DataFrame is empty.
        TypeError
            If the DataFrame has non-scalar value(s) (any iterable except string).
        ValueError
            If the DataFrame has duplicate rows.

        See Also
        --------
        pandas.Series.opti.to_indexset : Cast a Series into an IndexSet1D.
        pandas.Index.opti.to_indexset : Cast an Index into an IndexSet1D/IndexSetND.

        Examples
        --------
        >>> import pandas as pd
        >>> import opti_extensions  # required to access this method

        Casting a single-column DataFrame into an IndexSet1D

        >>> warehouses = pd.DataFrame(['WH-A', 'WH-B', 'WH-C'], columns=['WAREHOUSE'])
        >>> warehouses
          WAREHOUSE
        0      WH-A
        1      WH-B
        2      WH-C

        >>> warehouses.opti.to_indexset()
        IndexSet1D: (WAREHOUSE)
        ['WH-A', 'WH-B', 'WH-C']

        Casting a multi-column DataFrame into an IndexSetND

        >>> df = pd.DataFrame(
        ...     {
        ...         'PRODUCT': ['chair', 'chair', 'desk', 'desk'],
        ...         'WAREHOUSE': ['WH-A', 'WH-B', 'WH-B', 'WH-C'],
        ...         'PERIOD': [0, 1, 0, 1],
        ...     }
        ... )
        >>> df
          PRODUCT WAREHOUSE  PERIOD
        0   chair      WH-A       0
        1   chair      WH-B       1
        2    desk      WH-B       0
        3    desk      WH-C       1

        >>> df.opti.to_indexset()
        IndexSetND: (PRODUCT, WAREHOUSE, PERIOD)
        [('chair', 'WH-A', 0),
         ('chair', 'WH-B', 1),
         ('desk', 'WH-B', 0),
         ('desk', 'WH-C', 1)]

        >>> df.set_index('WAREHOUSE')[['PRODUCT', 'PERIOD']].opti.to_indexset()
        IndexSetND: (PRODUCT, PERIOD)
        [('chair', 0),
         ('chair', 1),
         ('desk', 0),
         ('desk', 1)]
        """
        if self._df.columns.size > 1:  # multi-column df
            vals_to_check = (
                self._df.select_dtypes(exclude=['number', 'datetime', 'datetimetz', 'timedelta'])
                .to_numpy()
                .flat
            )
            if any(isinstance(i, Iterable) and not isinstance(i, str) for i in vals_to_check):
                raise TypeError('DataFrame has non-scalar value(s) (no iterables except string)')

            names = (
                None
                if any(x is None for x in self._df.columns)
                else list(map(str, self._df.columns))
            )
            return IndexSetND(self._df.to_records(index=False).tolist(), names=names)

        else:  # single-column df
            return IndexSet1D(self._df.iloc[:, 0].tolist(), name=str(self._df.columns[0]))

    def to_paramdict(
        self,
    ) -> ParamDict1D[Any, int | float] | ParamDictND[tuple[Any, ...], int | float]:
        """Cast a single-column DataFrame into a ParamDict1D/ParamDictND.

        Note: The `opti-extensions` package has to be imported first to use this method with
        `pandas`.

        Returns
        -------
        ParamDict1D or ParamDictND
            * A single-index, single-column DataFrame will be cast into a ParamDict1D having index
              label as keys and column values as values. The index name will be set as the
              `ParamDict1D.key_name` attribute (as str). The column name will be set as the
              `ParamDict1D.value_name` attribute (as str).
            * A multi-index, single-column DataFrame will be cast into a ParamDictND having tuple of
              index labels as keys and column values as values. The multi-index names will be set as
              the `ParamDictND.key_names` attribute (as list[str]). The column name will be set as
              the `ParamDict1D.value_name` attribute (as str).

        Raises
        ------
        ValueError
            If the DataFrame is empty.
        ValueError
            If the DataFrame has multiple columns.
        ValueError
            If the DataFrame has duplicate index label(s).
        TypeError
            If the DataFrame has index label(s) that are not scalar (any iterable except string).
        TypeError
            If the DataFrame column values are not int or float.

        See Also
        --------
        pandas.Series.opti.to_paramdict : Cast a Series into a ParamDict1D/ParamDictND.

        Examples
        --------
        >>> import pandas as pd
        >>> import opti_extensions  # required to access this method

        Casting a single-index, single-column DataFrame into a ParamDict1D

        >>> inventory = pd.DataFrame({'PRODUCT': ['chair', 'desk'], 'UNITS': [200, 500]})
        >>> inventory
          PRODUCT  UNITS
        0   chair    200
        1    desk    500
        >>> inventory_units = inventory.set_index('PRODUCT')
        >>> inventory_units
                 UNITS
        PRODUCT
        chair      200
        desk       500

        >>> inventory_units.opti.to_paramdict()
        ParamDict1D: PRODUCT -> UNITS
        {'chair': 200, 'desk': 500}

        Casting a multi-index, single-column DataFrame into a ParamDictND

        >>> routes = pd.DataFrame(
        ...     {
        ...         'ORI': ['Delhi', 'Delhi', 'Seattle', 'Tokyo'],
        ...         'DES': ['Seattle', 'Tokyo', 'Tokyo', 'Delhi'],
        ...         'DIST': [11303, 5836, 7695, 5830],
        ...     }
        ... )
        >>> routes
               ORI      DES   DIST
        0    Delhi  Seattle  11303
        1    Delhi    Tokyo   5836
        2  Seattle    Tokyo   7695
        3    Tokyo    Delhi   5830
        >>> routes_dist = routes.set_index(['ORI', 'DES'])
        >>> routes_dist
                          DIST
        ORI     DES
        Delhi   Seattle  11303
                Tokyo     5836
        Seattle Tokyo     7695
        Tokyo   Delhi     5830

        >>> routes_dist.opti.to_paramdict()
        ParamDictND: (ORI, DES) -> DIST
        {('Delhi', 'Seattle'): 11303,
         ('Delhi', 'Tokyo'): 5836,
         ('Seattle', 'Tokyo'): 7695,
         ('Tokyo', 'Delhi'): 5830}
        """
        _check_duplicate_idx(self._df)

        if self._df.columns.size == 1:  # single-column df
            if self._df.index.nlevels > 1:  # multi-level index
                key_names = (
                    None
                    if any(x is None for x in self._df.index.names)
                    else list(map(str, self._df.index.names))
                )
                return ParamDictND(
                    self._df.iloc[:, 0].to_dict(),
                    key_names=key_names,
                    value_name=str(self._df.columns[0]),
                )

            else:  # single-level index
                key_name = None if self._df.index.name is None else str(self._df.index.name)
                return ParamDict1D(
                    self._df.iloc[:, 0].to_dict(),
                    key_name=key_name,
                    value_name=str(self._df.columns[0]),
                )

        else:  # multi-column df
            raise ValueError('`to_paramdict` does not support multi-column DataFrame')


class SeriesAccessor:
    """Accessor to cast pandas Series into IndexSet1D and ParamDict1D/ParamDictND.

    Parameters
    ----------
    series : Series
    """

    def __init__(self, series: Series[Any]):
        _check_empty(series)
        self._series = series

    def to_indexset(self) -> IndexSet1D[Any]:
        """Cast a Series into an IndexSet1D.

        Note: The `opti-extensions` package has to be imported first to use this method with
        `pandas`.

        Returns
        -------
        IndexSet1D
            * A Series will be cast into an IndexSet1D consisting of its values. The Series name
              will be set as the IndexSet1D.name attribute (as str).

        Raises
        ------
        ValueError
            If the Series is empty.
        TypeError
            If the Series has non-scalar value(s) (any iterable except string).
        ValueError
            If the Series has duplicate values.

        See Also
        --------
        pandas.DataFrame.opti.to_indexset : Cast a DataFrame into an IndexSet1D/IndexSetND.
        pandas.Index.opti.to_indexset : Cast an Index into an IndexSet1D/IndexSetND.

        Examples
        --------
        >>> import pandas as pd
        >>> import opti_extensions  # required to access this method

        Casting a Series into an IndexSet1D:

        >>> cities = pd.Series(['Delhi', 'Seattle', 'Tokyo'], name='CITY')
        >>> cities
        0      Delhi
        1    Seattle
        2      Tokyo
        Name: CITY, dtype: object

        >>> cities.opti.to_indexset()
        IndexSet1D: (CITY)
        ['Delhi', 'Seattle', 'Tokyo']
        """
        name = None if self._series.name is None else str(self._series.name)
        return IndexSet1D(self._series, name=name)

    def to_paramdict(
        self,
    ) -> ParamDict1D[Any, int | float] | ParamDictND[tuple[Any, ...], int | float]:
        """Cast a Series into a ParamDict1D/ParamDictND.

        Note: The `opti-extensions` package has to be imported first to use this method with
        `pandas`.

        Returns
        -------
        ParamDict1D or ParamDictND
            * A single-index Series will be cast into a ParamDict1D having index label as keys and
              Series values as values. The index name will be set as the `ParamDict1D.key_name`
              attribute (as str). The Series name will be set as the `ParamDict1D.value_name`
              attribute (as str).
            * A multi-index Series will be cast into a ParamDictND having tuple of index labels as
              keys and Series values as values. The multi-index names will be set as the
              `ParamDictND.key_names` attribute (as list[str]). The Series name will be set as the
              `ParamDict1D.value_name` attribute (as str).

        Raises
        ------
        ValueError
            If the Series is empty.
        ValueError
            If the Series has duplicate index label(s).
        TypeError
            If the Series has index label(s) that are not scalar (any iterable except string).
        TypeError
            If the Series values are not int or float.

        See Also
        --------
        pandas.DataFrame.opti.to_paramdict : Cast a single-column DataFrame into a ParamDict1D/
            ParamDictND.

        Examples
        --------
        >>> import pandas as pd
        >>> import opti_extensions  # required to access this method

        Casting a single-index Series into a ParamDict1D

        >>> inventory = pd.DataFrame({'PRODUCT': ['chair', 'desk'], 'UNITS': [200, 500]})
        >>> inventory
          PRODUCT  UNITS
        0   chair    200
        1    desk    500
        >>> inventory_units = inventory.set_index('PRODUCT')
        >>> inventory_units
                 UNITS
        PRODUCT
        chair      200
        desk       500

        >>> inventory_units.UNITS.opti.to_paramdict()
        ParamDict1D: PRODUCT -> UNITS
        {'chair': 200, 'desk': 500}

        Casting a multi-index Series into a ParamDictND

        >>> routes = pd.DataFrame(
        ...     {
        ...         'ORI': ['Delhi', 'Delhi', 'Seattle', 'Tokyo'],
        ...         'DES': ['Seattle', 'Tokyo', 'Tokyo', 'Delhi'],
        ...         'DIST': [11303, 5836, 7695, 5830],
        ...     }
        ... )
        >>> routes
               ORI      DES   DIST
        0    Delhi  Seattle  11303
        1    Delhi    Tokyo   5836
        2  Seattle    Tokyo   7695
        3    Tokyo    Delhi   5830
        >>> routes_dist = routes.set_index(['ORI', 'DES'])
        >>> routes_dist
                          DIST
        ORI     DES
        Delhi   Seattle  11303
                Tokyo     5836
        Seattle Tokyo     7695
        Tokyo   Delhi     5830

        >>> routes_dist.DIST.opti.to_paramdict()
        ParamDictND: (ORI, DES) -> DIST
        {('Delhi', 'Seattle'): 11303,
         ('Delhi', 'Tokyo'): 5836,
         ('Seattle', 'Tokyo'): 7695,
         ('Tokyo', 'Delhi'): 5830}
        """
        _check_duplicate_idx(self._series)

        value_name = None if self._series.name is None else str(self._series.name)

        if self._series.index.nlevels > 1:  # multi-level index
            key_names = (
                None
                if any(x is None for x in self._series.index.names)
                else list(map(str, self._series.index.names))
            )
            return ParamDictND(self._series.to_dict(), key_names=key_names, value_name=value_name)

        else:  # single-level index
            key_name = None if self._series.index.name is None else str(self._series.index.name)
            return ParamDict1D(self._series.to_dict(), key_name=key_name, value_name=value_name)


class IndexAccessor:
    """Accessor to cast pandas Index into IndexSet1D/IndexSetND.

    Parameters
    ----------
    idx : Index
    """

    def __init__(self, idx: Index[Any]):
        _check_empty(idx)
        self._idx = idx

    def to_indexset(self) -> IndexSet1D[Any] | IndexSetND[tuple[Any, ...]]:
        """Cast an Index into an IndexSet1D/IndexSetND.

        Note: The `opti-extensions` package has to be imported first to use this method with
        `pandas`.

        Returns
        -------
        IndexSet1D or IndexSetND
            * A single-level index will be cast into an IndexSet1D consisting of its values. The
              index name will be set as the `IndexSet1D.name` attribute (as str).
            * A multi-level index will be cast into an IndexSetND consisting of its values. The
              column names will be set as the `IndexSetND.names` attribute (as list[str]).

        Raises
        ------
        ValueError
            If the Index is empty.
        TypeError
            If the Index has non-scalar value(s) (any iterable except string).
        ValueError
            If the Index has duplicate values.

        See Also
        --------
        pandas.Series.opti.to_indexset : Cast a Series into an IndexSet1D.
        pandas.DataFrame.opti.to_indexset : Cast a DataFrame into an IndexSet1D/IndexSetND.

        Examples
        --------
        >>> import pandas as pd
        >>> import opti_extensions  # required to access this method

        Casting a DataFrame's single-level row index into an IndexSet1D:

        >>> warehouses = pd.DataFrame(
        ...     [10, 10, 5],
        ...     index=pd.Index(['WH-A', 'WH-B', 'WH-C'], name='WAREHOUSE'),
        ...     columns=['INVENTORY'],
        ... )
        >>> warehouses
                   INVENTORY
        WAREHOUSE
        WH-A              10
        WH-B              10
        WH-C               5

        >>> warehouses.index.opti.to_indexset()
        IndexSet1D: (WAREHOUSE)
        ['WH-A', 'WH-B', 'WH-C']

        Casting a DataFrame's multi-level row index into an IndexSetND:

        >>> warehouses = pd.DataFrame(
        ...     [10, 10, 5],
        ...     index=pd.MultiIndex.from_tuples(
        ...         [('EAST', 'WH-A'), ('EAST', 'WH-B'), ('WEST', 'WH-C')],
        ...         names=('ZONE', 'WAREHOUSE'),
        ...     ),
        ...     columns=['CAPACITY'],
        ... )
        >>> warehouses
                        CAPACITY
        ZONE WAREHOUSE
        EAST WH-A             10
             WH-B             10
        WEST WH-C              5

        >>> warehouses.index.opti.to_indexset()
        IndexSetND: (ZONE, WAREHOUSE)
        [('EAST', 'WH-A'), ('EAST', 'WH-B'), ('WEST', 'WH-C')]

        Casting a DataFrame's single-level column index into an IndexSet1D:

        >>> warehouses = pd.DataFrame(
        ...     {'WH-A': [10], 'WH-B': [10], 'WH-C': [5]}, index=['CAPACITY']
        ... ).rename_axis('WAREHOUSE', axis=1)
        >>> warehouses
        WAREHOUSE  WH-A  WH-B  WH-C
        CAPACITY     10    10     5

        >>> warehouses.columns.opti.to_indexset()
        IndexSet1D: (WAREHOUSE)
        ['WH-A', 'WH-B', 'WH-C']

        Casting a DataFrame's multi-level column index into an IndexSetND:

        >>> warehouses = pd.DataFrame(
        ...     {('EAST', 'WH-A'): [10], ('EAST', 'WH-B'): [10], ('WEST', 'WH-C'): [5]},
        ...     index=['CAPACITY'],
        ... ).rename_axis(('ZONE', 'WAREHOUSE'), axis=1)
        >>> warehouses
        ZONE      EAST      WEST
        WAREHOUSE WH-A WH-B WH-C
        CAPACITY    10   10    5

        >>> warehouses.columns.opti.to_indexset()
        IndexSetND: (ZONE, WAREHOUSE)
        [('EAST', 'WH-A'), ('EAST', 'WH-B'), ('WEST', 'WH-C')]

        Casting a Series' single-level index into an IndexSet1D:

        >>> cities = pd.Series(
        ...     ['A', 'B', 'C'],
        ...     index=pd.Index(['Delhi', 'Seattle', 'Tokyo'], name='CITY'),
        ...     name='ID',
        ... )
        >>> cities
        CITY
        Delhi      A
        Seattle    B
        Tokyo      C
        Name: ID, dtype: object

        >>> cities.index.opti.to_indexset()
        IndexSet1D: (CITY)
        ['Delhi', 'Seattle', 'Tokyo']

        Casting a Series' multi-level index into an IndexSetND:

        >>> cities = pd.Series(
        ...     ['A', 'B', 'C'],
        ...     index=(('Asia', 'Delhi'), ('Americas', 'Seattle'), ('Asia', 'Tokyo')),
        ...     name='ID',
        ... ).rename_axis(['ZONE', 'CITY'])
        >>> cities
        ZONE      CITY
        Asia      Delhi      A
        Americas  Seattle    B
        Asia      Tokyo      C
        Name: ID, dtype: object

        >>> cities.index.opti.to_indexset()
        IndexSetND: (ZONE, CITY)
        [('Asia', 'Delhi'), ('Americas', 'Seattle'), ('Asia', 'Tokyo')]
        """
        if self._idx.nlevels > 1:  # multi-level index
            names = (
                None if any(x is None for x in self._idx.names) else list(map(str, self._idx.names))
            )
            return IndexSetND(self._idx, names=names)

        else:
            name = None if self._idx.name is None else str(self._idx.name)
            return IndexSet1D(self._idx, name=name)
